// editor '93
// Copyright © 2009-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static bool
	aborting;		// keeps latched abort state

bool AbortCheck()
// See if the user is trying to abort an operation.
// NOTE: If an abort input is seen, this latches the abort state
// until later cleared by AbortClear. This is done so that if
// a Tcl script is within a "catch" function when the abort occurs
// (which would cause the script to ignore the error that is inserted
// during the abort), we'll keep telling the script to abort with each
// new command that is executed until it finally stops.
// This is a bit ugly, but Tcl does not provide an uncatchable error.
// Perhaps later, we could overload Tcl's catch function with one which
// will recognize an abort.
{
	if(aborting||PlatformEventsAbortCheck())
	{
		aborting=true;
	}
	return(aborting);
}

void AbortClear()
// Clear any pending abort.
{
	aborting=false;
	PlatformEventsAbortClear();
}

void AbortUnInit()
// Undo what AbortInit did.
{
}

bool AbortInit()
// Initialize abort handling.
{
	AbortClear();
	return(true);
}
