// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".


// global definitions and structures known by the editor

#define	VERSION	"2.0"
#define	EDITION	"r2"

#define	ABS(a)		 	((a)<0?(-(a)):(a))
#define	elementsof(a)	(sizeof(a)/sizeof(a[0]))

// Frequency that PlatformMonotonicTime, and timeouts use
#define	SECOND						1000

// defines needed to handle text
#define	LOAD_SAVE_BUFFER_SIZE		65536

#define	MAX_PATH_NAME_LENGTH		4096			// size in bytes maximum path name

#define	VIEW_MAX_STYLES				32				// maximum number of styles in a view

#define	VIEW_MAX_TAB_SIZE			255				// maximum tab size for a view

#define	MAX_FONT_NAME_LENGTH		1024			// size in bytes maximum composite font name

#define	SCROLL_FOLLOW_DELAY			(SECOND/100)	// how long to wait between scrolls when tracking pointer in views and lists

#define	HORIZONTAL_SCROLL_THRESHOLD	32				// number of "pixels" to scroll horizontally when using scroll arrows

// keyboard modifier bits

// shifts
enum
{
	EES_SHIFT=0,
	EES_CTRL=1,
	EES_WINDOW=2,
	EES_ALT=3,
	EES_CMD=4,
};

// masks
enum
{
	EEM_SHIFT=(1<<EES_SHIFT),
	EEM_CTRL=(1<<EES_CTRL),
	EEM_WINDOW=(1<<EES_WINDOW),
	EEM_ALT=(1<<EES_ALT),
	EEM_CMD=(1<<EES_CMD),
};


// view event types
enum
{
	VET_KEY_DOWN,				// saw key press within view
	VET_KEY_REPEAT,				// saw key repeat within view
	VET_VKEY_DOWN,				// saw virtual key press within view
	VET_VKEY_REPEAT,			// saw virtual key repeat within view
	VET_POSITION_VERTICAL,		// attempt to position the view to given absolute line
	VET_POSITION_HORIZONTAL,	// attempt to position the view to given absolute pixel
	VET_CLICK,					// initial click in view
	VET_CLICK_HOLD,				// send while click is being held in view
	VET_CLICK_RELEASE			// sent at end of tracked click in view
};

// view virtual keys
enum
{
	VVK_NOP,					// does nothing -- used to indicate no virtual key associated with key event
	VVK_SCROLL_UP,
	VVK_SCROLL_DOWN,
	VVK_DOCUMENT_PAGE_UP,
	VVK_DOCUMENT_PAGE_DOWN,
	VVK_SCROLL_LEFT,
	VVK_SCROLL_RIGHT,
	VVK_DOCUMENT_PAGE_LEFT,
	VVK_DOCUMENT_PAGE_RIGHT,
// keys that are normally found on a keyboard, but not clearly associated with an ASCII code
	VVK_LEFT_ARROW,
	VVK_RIGHT_ARROW,
	VVK_UP_ARROW,
	VVK_DOWN_ARROW,
	VVK_HOME,
	VVK_END,
	VVK_PAGE_UP,
	VVK_PAGE_DOWN,
	VVK_INSERT,
	VVK_BACK_SPACE,
	VVK_DELETE,
	VVK_HELP,
	VVK_UNDO,
	VVK_REDO,
	VVK_UNDO_TOGGLE,
	VVK_CUT,
	VVK_COPY,
	VVK_PASTE,
	VVK_SELECT_ALL,
	VVK_FIND,
	VVK_RETURN,
	VVK_TAB,
	VVK_ESC,
};

// menu relationships
enum
{
	MR_NEXT_SIBLING,			// next sibling
	MR_PREVIOUS_SIBLING,		// previous sibling
	MR_FIRST_CHILD,				// first child
	MR_LAST_CHILD				// last child
};

// ways to home view (this type is used horizontally and vertically)
enum
{
	HT_NONE,					// do not move
	HT_LENIENT,					// if already on view, do not move, if off view, move as little as possible to get on view
	HT_SEMI_STRICT,				// if already on view, do not move, if off view, move strict
	HT_STRICT					// move strict
};

// search types
enum
{
	ST_FIND,
	ST_FIND_ALL,
	ST_REPLACE,
	ST_REPLACE_ALL
};

typedef bool ABORT_TEST_FUNCTION();			// function passed to test for abort various time consuming operations

typedef	struct editorFile EDITOR_FILE;		// this type is hidden from the editor (the editor only needs pointers to it)
typedef struct editorMenu EDITOR_MENU;		// menu definition (actual structure hidden from the editor)
typedef struct editorTask EDITOR_TASK;		// task definition (actual structure hidden)
typedef struct editorFont EDITOR_FONT;		// font definition (actual structure hidden)
typedef struct platformViewInfo PLATFORM_VIEW_INFO;	// platform's local view information

typedef struct tokenlist					// token list definition
{
	const char
		*token;
	int
		tokenNum;
} TOKEN_LIST;

typedef struct variableBinding VARIABLE_BINDING;	// opaque variable binding

typedef struct variableTable
{
	VARIABLE_BINDING
		*variableBindingTable[256];		// array of heads of linked lists of variable bindings (crude hash table to get some speed when locating a binding)
	VARIABLE_BINDING
		*variableBindingListHead;		// head of list than runs through all variables
} VARIABLE_TABLE;

typedef struct
{
	int32_t
		x,y;
	uint32_t
		w,h;
} EDITOR_RECT;

typedef uint32_t	EDITOR_COLOR;		// ARGB (alpha is meaningful)

typedef struct textUniverse TEXT_UNIVERSE;	// opaque text universe
typedef struct selectionUniverse SELECTION_UNIVERSE;
typedef struct styleUniverse STYLE_UNIVERSE;
typedef struct syntaxInstance SYNTAX_INSTANCE;
typedef struct undoUniverse UNDO_UNIVERSE;
typedef struct viewColorScheme VIEW_COLOR_SCHEME;
typedef struct viewColorSchemeFonts VIEW_COLOR_SCHEME_FONTS;
typedef struct backgroundHandler BACKGROUND_HANDLER;
typedef struct editorWindow EDITOR_WINDOW;	// implementation specific window structure

typedef struct markVisibility
{
	struct editorMark				// mark that is visible in the view
		*parentMark;
	struct editorView				// view that the mark is visible in
		*parentView;
	struct markVisibility			// links through list of marks visible on a given view
		*nextMark,
		*prevMark;
	struct markVisibility			// links through list of views a given mark is visible in
		*nextView,
		*prevView;
	EDITOR_COLOR
		activeBackgroundColor,		// color to use when drawing background of visible mark in active view
		inactiveBackgroundColor;	// color to use when drawing background of visible mark in inactive view
} MARK_VISIBILITY;

typedef struct editorMark			// marks hanging off of editor universe
{
	char
		*markName;					// if NULL, then this mark is considered temporary
	SELECTION_UNIVERSE
		*selectionUniverse;			// the selection universe that defines this mark
	MARK_VISIBILITY
		*headVisibility,			// linked list of views this mark is visible in
		*tailVisibility;
	struct editorMark
		*nextMark;					// next mark in the list, NULL if no more
} EDITOR_MARK;

typedef struct editorBufferHandle
{
	struct editorBuffer
		*buffer;					// points to the buffer being held
	struct editorBufferHandle
		*nextHandle; 				// next in linked list of handles
} EDITOR_BUFFER_HANDLE;

typedef struct editorBuffer
{
	EDITOR_BUFFER_HANDLE
		*bufferHandles;				// when something needs to know when a buffer has been deleted, it can get a handle to it which the buffer code will clear on deletion
	bool
		shellBusy;					// used by the shell to make sure it cannot modify a buffer while it is in use by some previous shell command
	bool
		fromFile;					// tells if contents of this buffer represent an actual file
	char
		*contentName;				// name that describes the contents of the buffer (if fromFile=true, then this is full absolute pathname of file)
	TEXT_UNIVERSE
		*textUniverse;				// all the text
	UNDO_UNIVERSE
		*undoUniverse;				// undo/redo information
	SELECTION_UNIVERSE
		*selectionUniverse,			// selection and cursor information
		*mathSelectionUniverse,		// saved selection universe at the start of math operations
		*auxSelectionUniverse;		// aux selection and cursor information (allows tasks to write to universe at place other than cursor)
	EDITOR_MARK
		*marks;						// pointer to head of list of marks for this universe, NULL if none
	STYLE_UNIVERSE
		*styleUniverse;				// maintains style array
	SYNTAX_INSTANCE
		*syntaxInstance;			// keeps track of syntax highlighting (no syntax highlights if NULL)
	BACKGROUND_HANDLER
		*taskBackgroundHandler;		// function installed to manage task data when a task is running
	EDITOR_TASK
		*task;						// pointer to the task record for this buffer, or NULL if no task linked to this buffer
	struct editorView
		*firstView;					// head of linked list of views onto this buffer
	VARIABLE_TABLE
		variableTable;				// table of variables linked to this buffer
	// NOTE: the members below help keep track of various things during replacements
	bool
		replaceHaveRange;			// tells if replacement range encountered yet
	uint64_t
		replaceStartOffset,			// composite start offset in text
		replaceEndOffset;			// composite end offset in text (during the change)
	int64_t
		replaceNumBytes;			// number of bytes added/(subtracted) from text
	// NOTE: the members below help keep track of various things during selection (this is a bit of a mess and should be simplified)
	uint64_t
		anchorStartPosition,		// hold the ends of the selection while tracking (used in both normal and columnar selection tracking)
		anchorEndPosition;
	int64_t
		anchorX,					// holds the X opposite the mouse while columnar selection tracking
		columnarLeftX,				// keep left and right sides of a columnar selection so it can be added to intelligently
		columnarRightX;
	bool
		haveStartX,					// tell if desiredStartX/desiredEndX are valid
		haveEndX;
	int32_t
		desiredStartX,				// desired X position of the start of the selection when expanding/reducing, or of the cursor when no selection (used when moving the cursor up or down)
		desiredEndX;
	bool
		haveCurrentEnd;				// true if there is a current end of the selection being moved (used when extending the selection with the cursor keys)
	bool
		currentIsStart;				// true if the current end being moved is the start
	struct editorBuffer
		*nextBuffer;				// next buffer in linked list of buffers
} EDITOR_BUFFER;

typedef struct viewDescriptor		// describes a view to creation routines
{
	EDITOR_BUFFER
		*buffer;					// the buffer on which to view
	EDITOR_RECT
		rect;						// position and size within parent window
	bool
		active;						// tells if the view is created active or not
	bool
		useUtf8;					// if false, view is 8-bit ASCII, if true, view is UTF-8
	uint32_t
		tabSize;					// initial number of columns per tab
	const char
		*fontName;					// the name of the font to use for the view
	const char
		*colorSchemeName;			// the name of the color scheme to copy into the view
	void
		(*viewTextChangedVector)(struct editorView *view);		// this vector is called for the view after a text change
	void
		(*viewSelectionChangedVector)(struct editorView *view);	// this vector is called for the view after a selection change
	void
		(*viewPositionChangedVector)(struct editorView *view);	// this vector is called for the view after position change has been made
	void
		(*viewMetaChangedVector)(struct editorView *view);		// this vector is called for the view after view or buffer meta data changes
} EDITOR_VIEW_DESCRIPTOR;

// NOTE: it is possible to have multiple views of the same editor buffer
typedef struct editorView
{
	EDITOR_BUFFER
		*parentBuffer;				// the buffer on which this view is placed
	EDITOR_WINDOW
		*parentWindow;				// the window that contains this view
	struct editorView
		*nextBufferView;			// next view of the buffer that this view is on
	bool
		wantHome;					// used during buffer update, tells if given view should be homed
	bool
		active;						// if true, then view is displayed as active
	bool
		useUtf8;					// if false, view is 8-bit ASCII, if true, view is UTF-8
	uint32_t
		tabSize;					// number of columns per tab
	VIEW_COLOR_SCHEME
		*colorScheme;				// pointer to the color scheme for this view
	VIEW_COLOR_SCHEME_FONTS
		*colorSchemeFonts;			// loaded fonts associated with the color scheme
	void
		(*viewTextChangedVector)(struct editorView *view);		// this vector is called for the view after a text change
	void
		(*viewSelectionChangedVector)(struct editorView *view);	// this vector is called for the view after a selection change
	void
		(*viewPositionChangedVector)(struct editorView *view);	// this vector is called for the view after position change has been made
	void
		(*viewMetaChangedVector)(struct editorView *view);		// this vector is called for the view after view or buffer meta data changes
	MARK_VISIBILITY
		*headVisibility,			// linked list of marks visible in this view
		*tailVisibility;
	PLATFORM_VIEW_INFO
		*viewInfo;					// pointer to platform specific info about this view (gui needs to remember font, size, bounds, top line, etc)
} EDITOR_VIEW;

// The EDITOR_KEY structure is used for VET_KEY_DOWN, and VET_KEY_REPEAT events.
// modifiers are editor modifiers (not platform specific).
// keyCode is the platform specific (editor opaque) keyboard keycode.
typedef struct
{
	uint32_t
		modifiers;					// modifier bits (enumerated above)
	uint32_t
		keyCode;					// gives keyboard position of non-modifier key that was hit, or virtual editor key
	uint32_t
		vKeyCode;					// virtual key associated with the event (VVK_NOP if none)
	uint8_t
		numTextChars;				// tells how many text characters the key-press translated to
	uint8_t
		text[32];					// buffer holding the text characters
} EDITOR_KEY;

// The EDITOR_VKEY structure is used for VET_VKEY_DOWN, and VET_VKEY_REPEAT events.
// modifiers are editor modifiers (not platform specific).
// keyCode is the editor virtual key.
typedef struct
{
	uint32_t
		modifiers;					// modifier bits (enumerated above)
	uint32_t
		vKeyCode;					// virtual keycode
} EDITOR_VKEY;

typedef struct
{
	uint32_t
		modifiers;					// keyboard style modifier bits (enumerated above)
	uint64_t
		position;					// absolute position (in pixels, or lines)
} VIEW_POS_EVENT_DATA;

typedef struct
{
	uint32_t
		modifiers;					// keyboard style modifier bits (enumerated above)
	uint8_t
		button;						// mouse button (0-n)
	uint8_t
		numClicks;					// rapid click counter
	int32_t
		xClick,
		yClick;						// view relative click point
} VIEW_CLICK_EVENT_DATA;

typedef struct						// high level view events (sent to the shell from the view)
{
	uint16_t
		eventType;
	EDITOR_VIEW
		*view;						// the view that this event belongs to
	void
		*eventData;					// pointer to event specific data for event
} VIEW_EVENT;

