// High level document window stuff
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

EDITOR_WINDOW *EditorDocumentWindowLocate(char *bufferName)
// Attempt to locate a preexisting document window that
// is open to bufferName. If one is found, return it.
// If none is located, return NULL.
{
	EDITOR_BUFFER
		*buffer;

	if((buffer=EditorBufferLocate(bufferName)))
	{
		return(PlatformDocumentWindowLocate(buffer));
	}
	return(NULL);
}

void EditorDocumentWindowTitleReassign(EDITOR_BUFFER *buffer)
// If buffer has a window associated with it, assign the
// window's title based on buffer's contentName.
// If there is no window, do nothing.
{
	EDITOR_WINDOW
		*window;
	char
		*windowTitle;

	if((window=PlatformDocumentWindowLocate(buffer)))
	{
		if(buffer->fromFile)
		{
			if((windowTitle=PlatformWindowTitleFromPathCreate(buffer->contentName)))
			{
				PlatformDocumentWindowTitleSet(window,windowTitle);				// change the window's title
				MDisposePtr(windowTitle);										// get rid of memory
			}
			else
			{
				PlatformDocumentWindowTitleSet(window,buffer->contentName);		// had problem getting new name, use path
			}
		}
		else
		{
			PlatformDocumentWindowTitleSet(window,buffer->contentName);			// change the window's title
		}
	}
}

void EditorDocumentWindowClose(EDITOR_WINDOW *window)
// Close the given edit window (do not destroy the buffer it is tied to).
// This is not allowed to fail.
{
	PlatformDocumentWindowClose(window);										// get rid of the window
}

EDITOR_WINDOW *EditorDocumentWindowOpen(EDITOR_BUFFER *buffer,EDITOR_RECT *rect,bool useUtf8,uint32_t tabSize,const char *fontName,const char *colorSchemeName)
// Create a new document edit window for buffer, place it as the front-most.
// NOTE: if there is already a document window for buffer, it's considered
// an error.
// If there is a problem, ErrorSet, return NULL.
{
	EDITOR_WINDOW
		*window;
	char
		*titleToDelete,
		*windowTitle;

	window=NULL;
	if(!PlatformDocumentWindowLocate(buffer))
	{
		titleToDelete=NULL;
		if(buffer->fromFile)
		{
			titleToDelete=PlatformWindowTitleFromPathCreate(buffer->contentName);		// convert path name into more meaningful window title (name first)
		}
		if(titleToDelete)
		{
			windowTitle=titleToDelete;
		}
		else
		{
			windowTitle=buffer->contentName;		// just use buffer name as window title
		}
		window=PlatformDocumentWindowOpen(buffer,rect,windowTitle,useUtf8,tabSize,fontName,colorSchemeName);
		if(titleToDelete)
		{
			MDisposePtr(titleToDelete);
		}
	}
	else
	{
		ErrorSet("Window already open");
	}
	return(window);
}
