# e93 base startup file
# This is the top of the startup script tree for e93
# it finds this file, which in turn includes all other
# files which are executed at startup time.

# NOTE: normally, this file requires no modifications,
# since all of its functionality can be overridden in
# other files.

# WARNING: be careful when editing this file,
# e93 will stop processing this file at the FIRST error!

# Globals/Defaults
set			::env(TERM)					emacs;						# set environment variable to tell what kind of terminal this is

set			::tcl_precision				16;							# increase precision

set			::sysFilesDir				"[file dirname $SCRIPTPATH]";	# place where e93's system files are stored
set			::userFilesDir				"~/.e93";					# place where user specific files are stored

set			::docsDir					"documentation";			# place inside files directory where documenation lives
set			::mapsDir					"syntaxmaps";				# place inside files directory where syntaxmaps are stored
set			::auxMapsDir				"syntaxmaps_aux";			# user can add additional maps in here
set			::highlightDir				"highlightschemes";			# place inside files directory where highlighting schemes are stored
set			::auxHighlightDir			"highlightschemes_aux";		# user can add additional highlight schemes in here
set			::themesDir					"themes";					# place inside files directory where themes are stored
set			::auxThemesDir				"themes_aux";				# user can add additional themes in here
set			::modulesDir				"modules";					# place inside files directory where modules are stored
set			::auxModulesDir				"modules_aux";				# user can add additional modules in here
set			::prefsFile					"preferences.tcl";			# user preferences go here

set			::untitledCounter			0;							# used to give each new untitled window a unique name
set			::lineWrap					0;							# no line wrapping currently
set			::lineWrapColumn			70;							# column where lines should be wrapped

set			::windowWidthLimit			1400;						# width in pixels that we hold new windows to on really large screens

set			::grepFilesChoice			"";							# initialize the data in the grep files dialog
set			::replaceSelectionsChoice	"";							# initialize the data in the replace selections dialog
set			::pipeSelectionsChoice		"";							# initialize the data in the pipe selections dialog
set			::pipeWindowsChoice			"";							# initialize the data in the pipe windows dialog
set			::mailRecipientChoice		"";							# initialize the data in the mail recipient dialog
set			::mailSubjectChoice			"";							# initialize the data in the mail subject dialog
set			::refillLimitChoice			"70";						# initialize the data in the refill limit dialog

# Set the characters considered word breaks when double clicking.
setwordspacechars {0x00 0x2F} {0x3A 0x3F} {0x5B 0x5E} {0x60} {0x7B 0x7F} {0x80 0x9F} {0xA1 0xA9} {0xAB 0xB9} {0xBB 0xBF} {0xD7} {0xF7} {0x2000 0x206F} {0x20A0 0x20FF} {0x2190 0x26FF}

# Set checkboxes in the search dialog to default values.
# These will generally be overridden by user preferences, but exist here
# in case there are no preferences currently set up.
set			::sd_backwards				0;							# do not search backwards by default at startup
set			::sd_wrapAround				0;							# do not wrap around by default at startup
set			::sd_ignoreCase				0;							# do not ignore case by default
set			::sd_limitScope				0;							# do not limit scope by default
set			::sd_selectionExpression	0;							# do not use selection expressions by default at startup
set			::sd_useSubexpression		0;							# do not use subexpression by default at startup
set			::sd_subexpressionIndex		0;							# set subexpression index to 0 by default at startup
set			::sd_replaceProc			0;							# do not treat replace text as procedure

# if e93 is unable to match a document name, these will be used as the defaults
# These will generally be overridden by user preferences, but exist here
# in case there are no preferences currently set up.
set			::defaultUtf8				1;							# set to 1 to open new windows in UTF-8 mode, set to 0 for plain ASCII
set			::defaultTabSize			4
set			::defaultFontName			"liberation mono:size=9";
set			::defaultColorScheme		"e93";						# these must exist, or e93 will complain
set			::defaultSyntaxMap			"Plain Text";				# by default, we don't know what kind of document it is


# Styles enumerations for syntax coloring rules
# These style numbers are arbitrary
set			::style_default				0;
set			::style_comment				1;
set			::style_link				2;
set			::style_string				3;
set			::style_char				4;
set			::style_digit				5;
set			::style_operator			6;
set			::style_variable			7;
set			::style_value				8;
set			::style_delimiter			9;
set			::style_keyword				10;
set			::style_type				11;
set			::style_directive			12;
set			::style_function			13;
set			::style_label				14;
set			::style_local_label			15;
set			::style_bold				16;
set			::style_indent0				17;
set			::style_indent1				18;
set			::style_indent2				19;
set			::style_indent3				20;
set			::style_indent4				21;
set			::lastStyle					$style_indent4;	# set this to allow others to add more styles at the end

# define coloring rules for various languages

# see if the user has his own set of maps specified, if so, skip including the defaults, and get only his
set chosenDir [file join $sysFilesDir $mapsDir]
if {[file exists [file join $userFilesDir $mapsDir]]} \
{
	set chosenDir [file join $userFilesDir $mapsDir]
} 
foreach file [lsort [glob -nocomplain [file join $chosenDir *.tcl]]] \
{
	source $file;
}

# get user's additional syntax maps (if any)
if {[file exists [file join $userFilesDir $auxMapsDir]]} \
{
	foreach file [lsort [glob -nocomplain [file join $userFilesDir $auxMapsDir *.tcl]]] \
	{
		source $file;
	}
}

# see if the user has his own set of highlight schemes specified, if so, skip including the defaults, and get only his
set chosenDir [file join $sysFilesDir $highlightDir]
if {[file exists [file join $userFilesDir $highlightDir]]} \
{
	set chosenDir [file join $userFilesDir $highlightDir]
} 
foreach file [lsort [glob -nocomplain [file join $chosenDir *.tcl]]] \
{
	source $file;
}

# get user's additional highlighting schemes (if any)
if {[file exists [file join $userFilesDir $auxHighlightDir]]} \
{
	foreach file [lsort [glob -nocomplain [file join $userFilesDir $auxHighlightDir *.tcl]]] \
	{
		source $file;
	}
}

# Clipboard buffers
newbuffer	clip0; setbuffervariable clip0 lowpriority "";				# set the lowpriority variable which tells us to treat windows on this buffer specially
newbuffer	clip1; setbuffervariable clip1 lowpriority "";
newbuffer	clip2; setbuffervariable clip2 lowpriority "";
newbuffer	clip3; setbuffervariable clip3 lowpriority "";
newbuffer	clip4; setbuffervariable clip4 lowpriority "";
newbuffer	clip5; setbuffervariable clip5 lowpriority "";
newbuffer	clip6; setbuffervariable clip6 lowpriority "";
newbuffer	clip7; setbuffervariable clip7 lowpriority "";
newbuffer	clip8; setbuffervariable clip8 lowpriority "";
newbuffer	clip9; setbuffervariable clip9 lowpriority "";

setclipboard clip0;		# set this as the current clipboard

# Search buffers
newbuffer	findBuffer; setbuffervariable findBuffer lowpriority "";				# create default buffer to use for "find"
newbuffer	replaceBuffer; setbuffervariable replaceBuffer lowpriority "";			# create default buffer to use for "replace"
newbuffer	tempFindBuffer; setbuffervariable tempFindBuffer lowpriority "";		# these buffers are for commands which are implemented with search/replace, but do not want to overwrite the find/replaceBuffers
newbuffer	tempReplaceBuffer; setbuffervariable tempReplaceBuffer lowpriority "";

# ---------------------------------------------------------------------------------------------------------------------------------

# Determine screen and window sizes.
set			::screenWidth			[lindex [screensize] 0]
set			::screenHeight			[lindex [screensize] 1]
set			::windowHeight			[expr $::screenHeight*3/4];	# set the initial default height for new windows
set			::windowWidth			[expr $::windowHeight*4/3];	# set the initial default width for new windows (make 4:3 based on width)

# Limit the width on really large screens, since at some point, extra width isn't
# getting you much.
if {$::windowWidth>$::windowWidthLimit} \
{
	set			::windowWidth			$::windowWidthLimit;
}

# Determine how windows stagger when they are opened.
set			::staggerInitialX		[expr $::screenWidth/24];	# set the initial default X position for new windows
set			::staggerInitialY		[expr $::screenHeight/16];	# set the initial default Y position for new windows
set			::staggerIncrementX		[expr $::screenWidth/160];	# amount to increment by
set			::staggerIncrementY		[expr $::screenHeight/160];
set			::staggerMaxX			[expr $::screenWidth*3/24]
set			::staggerOpenX			$::staggerInitialX;			# set the current X position for new windows
set			::staggerOpenY			$::staggerInitialY;			# set the current Y position for new windows

# ---------------------------------------------------------------------------------------------------------------------------------
# Procedures

# Get a list of available themes.
proc ThemeList {} \
{
	set result ""
	if {[file exists [file join $::userFilesDir $::themesDir]]} \
	{
		foreach file [lsort [glob -nocomplain [file join $::userFilesDir $::themesDir *.tcl]]] \
		{
			lappend result [file rootname [file tail $file]];
		}
	} \
	else \
	{
		# use all system defined themes
		foreach file [lsort [glob -nocomplain [file join $::sysFilesDir $::themesDir *.tcl]]] \
		{
			lappend result [file rootname [file tail $file]];
		}
	}

	# get user's additional themes (if any)
	if {[file exists [file join $::userFilesDir $::auxThemesDir]]} \
	{
		foreach file [lsort [glob -nocomplain [file join $::userFilesDir $::auxThemesDir *.tcl]]] \
		{
			lappend result [file rootname [file tail $file]];
		}
	}
	return $result
}

# Set the theme to the one with the given name.
proc ThemeSet {name} \
{
	# If user has overriding themes directory search that
	if {[file exists [file join $::userFilesDir $::themesDir]]} \
	{
		set temp [file join $::userFilesDir $::themesDir $name.tcl]
		if {[file exists $temp]} \
		{
			set file $temp
		}
	} \
	else \
	{
		set temp [file join $::sysFilesDir $::themesDir $name.tcl]
		if {[file exists $temp]} \
		{
			set file $temp
		}
	}
	# If no file found, assume it must be in the aux directory
	if {![info exists file]} \
	{
		set file [file join $::userFilesDir $::auxThemesDir $name.tcl]
	}

	# If something found, use it
	if {[file exists $file]} \
	{
		source $file
	}
}

# Create a new list from the passed one where all elements of the
# passed list beginning with underscore have been removed.
proc FilterSystemElements {list} \
{
	foreach item $list \
	{
		if {[string range $item 0 0]!="_"} \
		{
			lappend result $item
		}
	}
	return $result
}

# Convenience function for setting marks in buffers
# this will reset the mark if it already exists, and create it if
# it does not already exist
proc SetMark {buffer mark} \
{
	openmark $buffer $mark;										# create new mark if none existed already
	copyselection $buffer -destmark $mark;						# copy selection into it
}

# Convenience function for locating marks in buffers
# this will reset the mark if it already exists, and create it if
# it does not already exist
proc GotoMark {buffer mark} \
{
	copyselection $buffer -sourcemark $mark;					# copy the mark over the buffer's selection
}

# Get just the font size of the given window.
proc GetFontSize {window} \
{
	return [lrange [fontnamesplit [getfont $window]] 1 1]
}

# Set just the font size of the given window.
# This reads the window's composite font, splits it, adjusts the size,
# joins it back together, and then sets it back in the window.
proc SetFontSize {window size} \
{
	# pin the size to something reasonable
	if {$size<1} \
	{
		beep
		set size 1
	} \
	elseif {$size>255} \
	{
		beep
		set size 255
	}
	set splitName [fontnamesplit [getfont $window]]
	setfont $window [fontnamejoin {*}[lreplace $splitName 1 1 $size]]
}

# Home a window to the start of selection (with arguments)
proc HomeWindowToSelectionStart {window args} \
{
	set position [lindex [getselectionends $window] 0];			# get selection start position
	homewindow $window $position $position {*}$args;			# home to start
}

# Home a window to the selection (with arguments)
# This tries to get the whole selection onto the view if possible.
proc HomeWindowToSelection {window args} \
{
	set position [getselectionends $window];					# get selection ends
	set startPosition [lindex $position 0];						# get start position
	set endPosition [lindex $position 1];						# get end position
	homewindow $window $startPosition $endPosition {*}$args;	# home
}

# Return true if the passed window name is one of the low-priority windows.
# A low priority window is one that e93 should not ask to save, even if it
# has been modified...
proc LowPriority {window} \
{
	expr {[catch {getbuffervariable $window lowpriority}]==0}
}

# When the interpreter does not understand a given command, send it to a shell.
# NOTE: no interactive commands can be executed this way, because Tcl would
# be suspended (waiting for the interactive command to finish). If Tcl is
# suspended, then e93 is also suspended (waiting for the Tcl command to finish)
# So there is no way that e93 could provide input to an interactive command.
# Interactive commands should be run as tasks.
#proc unknown args \
#{
#	return [uplevel exec $args]
#}

# NOTE: the code above has been replaced by this:
proc console {} {}
set tcl_interactive 1
# those 2 lines tell the tcl built-in unknown routine to behave as
# desired, and also get us auto loading and other useful features


# If there is no selection, select the current line
proc SelectLineWhenNoSelection {buffer} \
{
	set ends [getselectionends $buffer]
	set start [lindex $ends 0]
	set end [lindex $ends 1]
	if {$start==$end} \
	{
		selectline $buffer [lindex [positiontolineoffset $buffer $start] 0];	# select the line the cursor is on
	}
}

# Copy the selection in buffer to the clipboard passed.
# If there is no selection in buffer, then select the line the cursor is on.
proc SmartCopy {buffer clipboard} \
{
	SelectLineWhenNoSelection $buffer;
	copy $buffer $clipboard;
	flushundos $clipboard;							# kill any undo information, since we don't need to keep it for the clipboard
}

# Cut the selection in buffer to the clipboard passed.
# If there is no selection in buffer, then select the line the cursor is on.
proc SmartCut {buffer clipboard} \
{
	SelectLineWhenNoSelection $buffer;
	cut $buffer $clipboard;
	flushundos $clipboard;							# kill any undo information, since we don't need to keep it for the clipboard
}

# Replace the entire contents of a buffer with the passed text, and clear undos on the buffer.
# NOTE: it would be bad if this were called on a document that had been edited for hours and not
# saved! -- It would blow away the entire contents.
proc TextToBuffer {buffer text} \
{
	selectall $buffer;								# select everything, so insert will remove it
	insert $buffer $text;							# write over everything in buffer
	flushundos $buffer;								# get rid of any undo information for this buffer
}

# Move the stagger position to the next spot for a new staggered window.
proc UpdateStaggerPosition {} \
{
	uplevel #0 \
	{
		incr staggerOpenX $staggerIncrementX;
		incr staggerOpenY $staggerIncrementY;
		if {$staggerOpenX>$staggerMaxX} \
		{
			set staggerOpenX $staggerInitialX;
		 	set staggerOpenY $staggerInitialY
		}
	}
}

# Tile the windows on the display.
proc TileWindows {} \
{
	set horizontalNumber 4;								# number of windows we want across
	set verticalNumber 4;								# number we want down
	set horizontalBorder 30;							# number of pixels to leave around edges of screen
	set verticalBorder 30;
	set screendimensions [screensize]
	set width [lindex $screendimensions 0]
	set height [lindex $screendimensions 1]
	set windowWidth [expr ($width-($horizontalBorder*2))/$horizontalNumber];	# get the width of each individual window
	set windowHeight [expr ($height-($verticalBorder*2))/$verticalNumber];		# get the height of each individual window
	set horizontalIndex 0; set verticalIndex 0;			# init counters that tell us where we are
	foreach window [windowlist] \
	{
		if {![LowPriority $window]} \
		{
			setrect $window [expr $horizontalBorder+$horizontalIndex*$windowWidth] [expr $verticalBorder+$verticalIndex*$windowHeight] [expr $windowWidth-20] [expr $windowHeight-30]
			incr horizontalIndex
			if {$horizontalIndex>=$horizontalNumber} \
			{
				set horizontalIndex 0;
				incr verticalIndex;
				if {$verticalIndex>=$verticalNumber} \
				{
					set verticalIndex 0;
				}
			}
		}
	}
}

# Stack the windows on the display.
proc StackWindows {} \
{
	set ::staggerOpenX $::staggerInitialX;					# set the initial default X
	set ::staggerOpenY $::staggerInitialY;					# set the initial default Y
	foreach window [lreverse [windowlist]] \
	{
		if {![LowPriority $window]} \
		{
			setrect $window $::staggerOpenX $::staggerOpenY $::windowWidth $::windowHeight
			UpdateStaggerPosition
		}
	}
}

# Work out the default tab size, color scheme, and highlight mode for buffer.
# Return a list which contains a boolean that tells if something other than the
# default is being returned, the encoding, tab size, the color scheme name, and the syntax map name.
proc GetDefaultStyleInfo {buffer} \
{
	set hadMatch 0

	# set up defaults
	set useUtf8 $::defaultUtf8
	set tabSize $::defaultTabSize
	set fontName $::defaultFontName
	set schemeName $::defaultColorScheme
	set mapName $::defaultSyntaxMap

	foreach extensionAttrName [array names ::extensionHuntExpression] \
	{
		if {[regexp $::extensionHuntExpression($extensionAttrName) [file tail $buffer]]} \
		{
			set useUtf8 $::extensionUtf8($extensionAttrName)
			set tabSize $::extensionTabSize($extensionAttrName)
			# font name is not taken from the extension...
			set schemeName $::extensionColorScheme($extensionAttrName)
			set mapName $::extensionMapName($extensionAttrName)
			set hadMatch 1
			break
		}
	}
	lappend temp $hadMatch $useUtf8 $tabSize $fontName $schemeName $mapName;		# return the items
}

# When a window's name has been changed, this is called to update the
# styles in use for the window.
# If the name is recognized, this will change the style, otherwise it
# does nothing.
proc UpdateStyle {window} \
{
	set temp [GetDefaultStyleInfo $window]
	if {[lindex $temp 0]} \
	{
		setutf8 $window [lindex $temp 1]
		settabsize $window [lindex $temp 2]
		# font is not updated when window name changes
		setcolorscheme $window [lindex $temp 4]
		setsyntaxmap $window [lindex $temp 5]
	}
}

# Create a window onto buffer, using given values for position, size, UTF-8 encoding, tabSize, font name, color scheme, and syntax map.
proc OpenWindow {buffer x y width height useUtf8 tabSize fontName schemeName mapName} \
{
	if {[haswindow $buffer]} \
	{
		settopwindow $buffer;						# if it was already open, then just put it to the top, change nothing
	} \
	else \
	{
		openwindow $buffer $x $y $width $height $useUtf8 $tabSize $fontName $schemeName
		setsyntaxmap $buffer $mapName
		HomeWindowToSelectionStart $buffer;			# go to the cursor/selection
	}
}

# Create a window onto buffer, using default values for position and size
# and given values for everything else.
proc OpenStaggeredWindow {buffer useUtf8 tabSize fontName schemeName mapName} \
{
	if {[haswindow $buffer]} \
	{
		settopwindow $buffer;						# if it was already open, then just put it to the top
	} \
	else \
	{
		if {[LowPriority $buffer]} \
		{
			set width [expr $::screenWidth*3/4]
			set height [expr $::screenHeight*1/8]
			set x [expr ($::screenWidth-$width)/2]
			set y [expr $::screenHeight-$height]
		} \
		else \
		{
			set width $::windowWidth
			set height $::windowHeight
			set x $::staggerOpenX
			set y $::staggerOpenY
			UpdateStaggerPosition;
		}
		OpenWindow $buffer $x $y $width $height $useUtf8 $tabSize $fontName $schemeName $mapName
	}
}

# Create a window onto buffer, using default values for position, size, tabSize, font name, color scheme, and highlighting.
proc OpenDefaultWindow {buffer} \
{
	if {[haswindow $buffer]} \
	{
		settopwindow $buffer;						# if it was already open, then just put it to the top
	} \
	else \
	{
		if {[LowPriority $buffer]} \
		{
			set useUtf8 1
			set tabSize 4
			set fontName $::defaultFontName
			set schemeName "Low Priority"
			set mapName ""
		} \
		else \
		{
			set temp [GetDefaultStyleInfo $buffer]
			set useUtf8 [lindex $temp 1]
			set tabSize [lindex $temp 2]
			set fontName [lindex $temp 3]
			set schemeName [lindex $temp 4]
			set mapName [lindex $temp 5]
		}
		OpenStaggeredWindow $buffer $useUtf8 $tabSize $fontName $schemeName $mapName
	}
}

# Return a good name for a new buffer.
proc NewUntitledBufferName {} \
{
	set name "Untitled-$::untitledCounter";			# get new name
	incr ::untitledCounter;							# next time, make the count one larger
	return $name;									# return the name of the new window
}

# Return a name as close to givenName as possible, avoiding duplicate
# buffer names.
proc NewTitledBufferName {givenName} \
{
	set count 0
	set name $givenName;							# get new name
	while {[bufferexists $name]} \
	{
		set name "$givenName-$count";
		incr count;
	}
	return $name;									# return the name of the new window
}

# Create a new window with a default name, return the name.
proc NewUntitledWindow {} \
{
	OpenDefaultWindow [set name [newbuffer [NewUntitledBufferName]]];	# open a new window with new name
	return $name;									# return the name of the new window
}

# Create a new window with a name as close as possible to the given name, return the name.
proc NewTitledWindow {givenName} \
{
	OpenDefaultWindow [set name [newbuffer [NewTitledBufferName $givenName]]];	# open a new window with new name
	return $name;									# return the name of the new window
}

# Open the passed file name into a buffer, and record the modification time
# of the file into the variable "mtime" attached to the buffer
# If there is a problem, fail just like openbuffer would.
proc OpenBufferRecordMtime {file} \
{
	if {[catch {openbuffer $file} message]==0} \
	{
		# see if there is ALREADY a modification time tag associated with this file, if so, check against current mtime and complain if they differ
		# NOTE: there could already be a variable assigned if the file was open already
		if {[catch {getbuffervariable $message mtime} oldmtime]==0} \
		{
			if {[catch {file mtime $message} newmtime]==0} \
			{
				if {$oldmtime!=$newmtime} \
				{
					okdialog "WARNING!\n\nIt looks like another application modified:\n$message\nwhile it was open for editing.\n"
				}
			} \
			else \
			{
				okdialog "WARNING!\n\nA buffer exists for the requested file:\n$message\nBut could not read mtime.\n"
			}
		} \
		else \
		{
			catch {setbuffervariable $message mtime [file mtime $message]};
		}
		return $message;
	} \
	else \
	{
		return -code error $message;
	}
}

# Open the passed file name into a buffer, and record the modification time
# of the file into the variable "mtime" attached to the buffer.
# Then open a default window onto the buffer.
# if there is a problem, fail just like openbuffer would.
proc OpenDefaultWindowRecordMtime {file} \
{
	if {[catch {OpenDefaultWindow [OpenBufferRecordMtime $file]} message]==0} \
	{
		return $message;
	} \
	else \
	{
		return -code error $message;
	}
}

# Attempt to open a list of files one at a time.
# If any fails to open, report why, and give the user a chance to cancel.
proc OpenList {list} \
{
	set numItems [llength $list]
	set item 0
	while {$item<$numItems} \
	{
		set file [lindex $list $item]
		incr item
		if {[catch {OpenDefaultWindowRecordMtime $file} errorMessage]!=0} \
		{
			if {$item==$numItems} \
			{
				okdialog "Failed to open '$file'\n$errorMessage"
			} \
			else \
			{
				okcanceldialog "Failed to open '$file'\n$errorMessage\n\nContinue?"
			}
		}
	}
}

# Attempt to open the given list of files.
# Return 1 if the user didn't cancel.
# Return 2 if the user canceled.
# This is similar to OpenList, but works for the SmartOpen functions
proc SmartOpenFileList {list line offset moreItems} \
{
	set numItems [llength $list]
	set item 0
	while {$item<$numItems} \
	{
		set file [lindex $list $item]
		incr item
		if {[catch {OpenDefaultWindow [set newBuffer [OpenBufferRecordMtime $file]]} message]==0} \
		{
			if {[string length $line] > 0} \
			{
				selectline $newBuffer $line;
				HomeWindowToSelectionStart $newBuffer;
				if {[string length $offset] > 0} \
				{
					if {$offset>0} \
					{
						set offset [expr $offset-1];		# decrement offset, since GNU tools count starting at 1, and e93 counts starting at 0
					}
					set position [lineoffsettoposition $newBuffer $line $offset]
					openmark $newBuffer _temp
					setselectionends $newBuffer $position [expr $position+1] -mark _temp
					markvisible $newBuffer _temp FF0000
				}
			}
			if {$item==$numItems} \
			{
				return 1;
			}
		} \
		else \
		{
			if {($item!=$numItems)||$moreItems} \
			{
				if {[catch {okcanceldialog "Failed to open '$file'\n$message\n\nContinue?"}]==0} \
				{
					if {$item==$numItems} \
					{
						return 1;
					}
				} \
				else \
				{
					return 2;
				}
			} \
			else \
			{
				okdialog "Failed to open '$file'\n$message"
				return 1;
			}
		}
	}
	return 1
}

# Given an item from a list, try to interpret it as the output from
# an GNU tool:
# filename:line
# or
# filename:line:offset
# If it looks like that, then extract the filename, line number and
# offset, and try to open it.
# If there's a problem opening the file, complain
proc SmartOpenGnu {listItem moreItems} \
{
	if {[regexp {^([^:]+):([0-9]+):([0-9]+):} $listItem whole file line offset]} \
	{
		return [SmartOpenFileList $file $line $offset $moreItems]
	}
	if {[regexp {^([^:]+):([0-9]+):} $listItem whole file line]} \
	{
		return [SmartOpenFileList $file $line "" $moreItems]
	}
	return 0;				# didn't recognize it
}

# Given an item from a list, try to interpret it as the output from
# an MPW tool:
# File filename;Line nnnn
# If it looks like that, then extract the filename and line number
# and try to open it, otherwise return an error
proc SmartOpenMpw {listItem moreItems} \
{
	if {[regexp {^File *'?([^' ]+)'? *; *Line *([0-9]+)} $listItem whole file line]} \
	{
		return [SmartOpenFileList $file $line "" $moreItems]
	}
	return 0;				# didn't recognize it
}

# Given an item from a list, try to interpret it as the output from
# a Visual Studio tool:
# filename(nnnn)
# If it looks like that, then extract the filename and line number
# and try to open it, otherwise return an error
proc SmartOpenVisualStudio {listItem moreItems} \
{
	if {[regexp {^([^ (]+) *\(([0-9]+)\)} $listItem whole file line]} \
	{
		return [SmartOpenFileList $file $line "" $moreItems]
	}
	return 0;				# didn't recognize it
}

# Given an item from a list, try to interpret it as a 'C' include
# statement:
# #include <filename>	(looks in /usr/include)
# #include "filename"	(looks in current directory)
# If it looks like that, then extract the filename
# and try to open it, otherwise return an error
proc SmartOpenInclude {listItem moreItems} \
{
	if {[regexp "^\[ \t\]*#\[ \t\]*include\[ \t\]+\\<(\[^>\]+)\\>" $listItem whole file]} \
	{
		return [SmartOpenFileList "/usr/include/$file" "" "" $moreItems]
	} \
	else \
	{
		if {[regexp "^\[ \t\]*#\[ \t\]*include\[ \t\]+\"(\[^\"\]+)\"" $listItem whole file]} \
		{
			return [SmartOpenFileList "$file" "" "" $moreItems]
		}
	}
	return 0;				# didn't recognize it
}

# Given an item from a list, try to interpret it as a URL.
# If it looks like that, then attempt to open the URL
# in the user's default browser.
proc SmartOpenUrl {listItem moreItems} \
{
	if {[regexp {^[A-Za-z]+[0-9A-Za-z+.\\-]*://.+} $listItem whole]} \
	{
		# handle file URLs directly
		if {[regexp -nocase {^file://(.+)} $listItem whole filePart]} \
		{
			if {[catch {OpenDefaultWindowRecordMtime $filePart} errorMessage]!=0} \
			{
				okdialog "Failed to open '$filePart'\n$errorMessage"
			}
		} \
		else \
		{
			exec xdg-open $whole &
		}
		return 1
	}
	return 0
}

# Given an item from a list, try to interpret it as a globbed list
# of names.
# If it looks like that, then extract the filenames and try
# to open them, otherwise return 0
proc SmartOpenGlob {listItem moreItems} \
{
	if {[catch {lsort [glob $listItem]} globList]==0} \
	{
		return [SmartOpenFileList $globList "" "" $moreItems]
	}
	return 0
}

# Just try to open the item as a file.
proc SmartOpenFile {listItem moreItems} \
{
	return [SmartOpenFileList [list $listItem] "" "" $moreItems]
}

# Passed a buffer, look at the selections within it, and work out what
# to open. This will open files if it can find them, or if it recognizes
# URLs, it will call to open those.
# NOTE: this also recognizes error output from various tools, and can
# open files to a given line.
proc SmartOpen {buffer} \
{
	set list [selectedtextlist $buffer]
	set numItems [llength $list]

	# if there are no selections, then see if the cursor is in a "link" style
	# if it is, select that, otherwise select the whole line the cursor is on
	if {$numItems==0} \
	{
		set ends [getselectionends $buffer]
		set start [lindex $ends 0]
		set styleInfo [getstyleatposition $buffer $start]
		if {[lindex $styleInfo 0]==$::style_link} \
		{
			setselectionends $buffer [lindex $styleInfo 1] [lindex $styleInfo 2]
		} \
		else \
		{
			SelectLineWhenNoSelection $buffer
		}
		set list [selectedtextlist $buffer]
		set numItems [llength $list]
	}

	set item 0
	while {$item<$numItems} \
	{
		set listItem [lindex $list $item]
		incr item

		# break lines into list items
		set listItem [string trimright $listItem "\n"];		# get rid of any trailing newlines
		set subList [split $listItem "\n"];					# break remaining newlines into separate items
		set numSubItems [llength $subList]
		set subItem 0

		while {$subItem<$numSubItems} \
		{
			set subListItem [lindex $subList $subItem]
			incr subItem

			foreach procListElement {SmartOpenGnu SmartOpenMpw SmartOpenVisualStudio SmartOpenInclude SmartOpenUrl SmartOpenGlob SmartOpenFile} \
			{
				# see if one of the functions will claim this item as its own
				set status [$procListElement $subListItem [expr ($item<$numItems)||($subItem<$numSubItems)]]
				switch $status \
				{
					"0"
					{
					}
					"1"
					{
						break;						# item was claimed
					}
					"2"
					{
						set subItem $numSubItems;	# force inner loop to terminate
						set item $numItems;			# force outer loop to terminate
						break;
					}
				}
			}
		}
	}
}

# Attempt to include a list of files one at a time, into the given buffer.
# If any fails to include, report why, and give the user a chance to cancel.
proc IncludeList {buffer list} \
{
	set numItems [llength $list]
	set item 0

	while {$item<$numItems} \
	{
		set file [lindex $list $item]
		incr item
		if {[catch {insertfile $buffer $file} errorMessage]!=0} \
		{
			if {$item==$numItems} \
			{
				okdialog "Failed to include '$file'\n$errorMessage"
			} \
			else \
			{
				okcanceldialog "Failed to include '$file'\n$errorMessage\n\nContinue?"
			}
		}
	}
}

# Swap the top two windows on the screen.
proc SwapWindows {} \
{
	set windowList [windowlist]
	if {[llength $windowList]>1} \
	{
		settopwindow [lindex $windowList 1]
	} \
	elseif {[llength $windowList]>0} \
	{
		settopwindow [lindex $windowList 0]
	} \
	else \
	{
		beep
	}
}

# Bring the bottom window to the top.
proc RotateWindows {} \
{
	set windowList [windowlist]
	set numWindows [llength $windowList]
	if {$numWindows>1} \
	{
		incr numWindows -1
		settopwindow [lindex $windowList $numWindows]
	} \
	elseif {$numWindows>0} \
	{
		settopwindow [lindex $windowList 0]
	} \
	else \
	{
		beep
	}
}

# Choose a window to be top-most (this places a * next to windows which have
# been modified)
proc ChooseWindow {} \
{
	set newList "";
	foreach window [windowlist]\
	{
		if {[isdirty $window]} \
		{
			lappend newList "* $window";
		} \
		else \
		{
			lappend newList "  $window";
		}
 	}
	foreach window [listdialog "Choose a Window:" $newList] \
	{
		settopwindow [string range $window 2 end];	# select the window (removing the modified indication)
	}
}

# Return the active window if there is one, otherwise
# just beep, and return an error.
proc ActiveWindowOrBeep {} \
{
	if {[catch {activewindow} message]==0} \
	{
		settopwindow $message;						# make sure the window is on the top (can be defocussed by the menu)
		return $message
	}
	beep
	return -code error
}

# Return the current clipboard if there is one, otherwise
# just beep, and return an error.
proc CurrentClipboardOrBeep {} \
{
	if {[catch {getclipboard} message]==0} \
	{
		 return $message
	}
	beep
	return -code error
}

# Report the current clipboard, or a message that indicates that
# there is none
proc ShowCurrentClipboard {} \
{
	if {[catch {getclipboard} message]==0} \
	{
		okdialog "Current clipboard:\n\n$message"
	} \
	else \
	{
		okdialog "There is no current clipboard"
	}
}

# Quiz user to find out if he really wants to "revert".
proc AskRevert {buffer} \
{
	if {[isdirty $buffer]} \
	{
		okcanceldialog "Do you really want to discard changes to:\n'$buffer'"
		revertbuffer $buffer
		if {[fromfile $buffer]} \
		{
			catch {setbuffervariable $buffer mtime [file mtime $buffer]};		# update the modification time so we can look at it when saving
		}
	} \
	else \
	{
		revertbuffer $buffer;			# may be reverting because the file was changed, so just do it without asking
		if {[fromfile $buffer]} \
		{
			catch {setbuffervariable $buffer mtime [file mtime $buffer]};		# update the modification time so we can look at it when saving
		}
	}
}

# Get name of file to "save as", then do it for the given buffer.
# NOTE: this returns the new name of the buffer if it completes successfully.
proc AskSaveAs {buffer} \
{
	set newPath [savedialog "Save File:" $buffer]
	if {[file exists $newPath]} \
	{
		okcanceldialog "File:\n'$newPath'\nalready exists. Overwrite?"
	}
	if {[catch {savebufferas $buffer $newPath} message]==0} \
	{
		catch {setbuffervariable $message mtime [file mtime $message]};		# update the modification time so we can look at it when saving
	} \
	else \
	{
		okdialog "Failed to save:\n'$newPath'\n$message"
		return -code error
	}
	UpdateStyle $message;				# update the styles of this window
	return $message;					# return the new name
}

# Get name of file to "save to", then do it for the given buffer.
proc AskSaveTo {buffer} \
{
	set newPath [savedialog "Save Copy To:" $buffer]
	if {[file exists $newPath]} \
	{
		okcanceldialog "File:\n'$newPath'\nalready exists. Overwrite?"
	}
	if {[catch {savebufferto $buffer $newPath} errorMessage]!=0} \
	{
		okdialog "Failed to save:\n'$newPath'\n$errorMessage"
		return -code error
	}
}

# See if buffer is not linked to a file, and if not, ask for
# a file name to save it to, else just save it.
# Either way, return the name of the buffer after it is saved,
# as it may have changed during the save process.
proc AskSave {buffer} \
{
	if {[fromfile $buffer]} \
	{
		# see if there is a modification time tag associated with this file, if so, check against current mtime and complain if they differ
		if {[catch {getbuffervariable $buffer mtime} oldmtime]==0} \
		{
			if {[catch {file mtime $buffer} newmtime]==0} \
			{
				if {$oldmtime!=$newmtime} \
				{
					okcanceldialog "WARNING!\n\nIt looks like another application modified:\n$buffer\nwhile it was open for editing.\n\nWould you like to save anyway?"
				}
			} \
			else \
			{
				okcanceldialog "WARNING!\n\nIt looks like another application deleted:\n$buffer\nwhile it was open for editing.\n\nWould you like to save anyway?"
			}
		}

		if {[catch {savebuffer $buffer} message]==0} \
		{
			catch {closemark $buffer _temp};									# clear any temporary mark
			catch {setbuffervariable $buffer mtime [file mtime $buffer]};		# update the modification time so we can look at it next time we save
			if {[catch {getbuffervariable $buffer aftersave} aftersave]==0} \
			{
				{*}$aftersave;													# run commands after buffer is saved
			}
		} \
		else \
		{
			okdialog "Failed to save:\n'$buffer'\n$message"
			return -code error
		}
	} \
	else \
	{
		set buffer [AskSaveAs $buffer];						# pick up the new name of the buffer
	}
	return $buffer
}

# Ask the user if he really wants to save all the dirty windows, if so, do it.
proc AskSaveAll {} \
{
	ActiveWindowOrBeep;										# make sure there is a window, if not, just beep and leave
	okcanceldialog "Really save all modified windows?"
	foreach window [windowlist] \
	{
		if {[isdirty $window]} \
		{
			if {[catch {AskSave $window}]!=0} \
			{
				okdialog "Save All aborted\n"
				return -code error
			}
		}
	}
}

# See if buffer is dirty before closing.
# If it is, give user a chance to save.
proc AskClose {buffer} \
{
	# See if buffer points to a something low-priority, we just close the windows on them, not the buffer
	if {[LowPriority $buffer]} \
	{
		closewindow $buffer
	} \
	else \
	{
		if {[isdirty $buffer]} \
		{
			settopwindow $buffer;							# pop it to the top of the window stack

			if {[yesnodialog "Save Changes To:\n'$buffer'\nBefore Closing?"]} \
			{
				set buffer [AskSave $buffer];				# if it is saved, get the new name so we can close it
			}
		}
		closebuffer $buffer
	}
}

# Ask for a list of windows which should be printed
proc PrintWindows {} \
{
	foreach buffer [listdialog "Choose Window(s) to Print:" [windowlist]] \
	{
		set command [textdialog "If needed, modify the print command below:" "[eval format "%s" \"$::defaultPrintCommand\"]"]
		SetMark $buffer temp;						# remember what was selected, we will not disturb it
		selectall $buffer;							# make a selection of all of the text in the passed buffer
		if {[catch {eval exec $command << {[lindex [selectedtextlist $buffer] 0]}} result]!=0} \
		{
			okdialog "Print status:\n\n$result\n"
		}
		GotoMark $buffer temp;						# put selection back
		closemark $buffer temp;						# get rid of temp selection
	}
}

# Attempt to get lpr to output this document to the local printer
proc PrintBuffer {buffer} \
{
	set command [textdialog "If needed, modify the print command below:" "[eval format "%s" \"$::defaultPrintCommand\"]"]
	SetMark $buffer temp;							# remember what was selected, we will not disturb it
	selectall $buffer;								# make a selection of all of the text in the passed buffer
	if {[catch {eval exec $command << {[lindex [selectedtextlist $buffer] 0]}} result]!=0} \
	{
		okdialog "Print status:\n\n$result\n"
	}
	GotoMark $buffer temp;							# put selection back
	closemark $buffer temp;							# get rid of temp selection
}

# print the first selection to the local printer
proc PrintSelection {buffer} \
{
	if {[lindex [selectioninfo $buffer] 6]} \
	{
		set command [textdialog "If needed, modify the print command below:" "[eval format "%s" \"$::defaultPrintCommand\"]"]
		if {[catch {eval exec $command << {[lindex [selectedtextlist $buffer] 0]}} result]!=0} \
		{
			okdialog "Print status:\n\n$result\n"
		}
	} \
	else \
	{
		beep
	}
}

# Attempt to close all open windows, asking to save any that are dirty.
# If the user cancels, then back out, otherwise, quit.
proc TryToQuit {} \
{
	foreach window [windowlist] \
	{
		AskClose $window
	}
	forceQUIT
}

# When nothing is found, look to see if -limitscope was set in the arguments, and
# throw up a dialog reminding the user (this can avoid confusion if -limitscope
# is accidentally left on)
proc FoundNothing {args} \
{
	if {[lsearch -exact {*}$args -limitscope]>=0} \
	{
		beep;
		okdialog "Limited scope search found nothing"
	} \
	else \
	{
		beep;
	}
}

# Just like find, but will place the result in a dialog if it fails, and home if it succeeds, or beep if it finds nothing.
proc FindBeep {window buffer args} \
{
	if {[catch {eval [list find $window $buffer] $args} message]==0} \
	{
		if {[llength $message]} {homewindow $window [lindex $message 0] [lindex $message 1]} else {FoundNothing $args}
	} \
	else \
	{
		okdialog $message
	}
}

# Just like findall, but will place the result in a dialog if it fails, and home if it succeeds, or beep if it finds nothing.
proc FindAllBeep {window buffer args} \
{
	if {[catch {eval [list findall $window $buffer] $args} message]==0} \
	{
		if {[llength $message]} {homewindow $window [lindex $message 0] [lindex $message 1]} else {FoundNothing $args}
	} \
	else \
	{
		okdialog $message
	}
}

# Just like replace, but will place the result in a dialog if it fails, and home if it succeeds, or beep if it finds nothing.
proc ReplaceBeep {window findBuffer replaceBuffer args} \
{
	if {[catch {eval [list replace $window $findBuffer $replaceBuffer] $args} message]==0} \
	{
		if {[llength $message]} {homewindow $window [lindex $message 0] [lindex $message 1]} else {FoundNothing $args}
	} \
	else \
	{
		okdialog $message
	}
}

# Just like replaceall, but will place the result in a dialog if it fails, and home if it succeeds, or beep if it finds nothing.
proc ReplaceAllBeep {window findBuffer replaceBuffer args} \
{
	if {[catch {eval [list replaceall $window $findBuffer $replaceBuffer] $args} message]==0} \
	{
		if {[llength $message]} {homewindow $window [lindex $message 0] [lindex $message 1]} else {FoundNothing $args}
	} \
	else \
	{
		okdialog $message
	}
}

# Turn the booleans that come back from the search dialog into flags which can be passed
# to the search functions
proc CreateSearchOptions {window} \
{
	set result ""

	# If the window is in UTF8 mode, then search in UTF8 mode
	if {[getutf8 $window]} \
	{
		lappend result "-utf8";
	}

	# Append options from search dialog
	if {$::sd_backwards!=0} \
	{
		lappend result "-backward";
	}
	if {$::sd_wrapAround!=0} \
	{
		lappend result "-wrap";
	}
	if {$::sd_ignoreCase!=0} \
	{
		lappend result "-ignorecase";
	}
	if {$::sd_limitScope!=0} \
	{
		lappend result "-limitscope";
	}
	if {$::sd_selectionExpression!=0} \
	{
		lappend result "-regex";
	}
	if {$::sd_useSubexpression!=0} \
	{
		lappend result "-usesubexpression";
		lappend result $::sd_subexpressionIndex;
	}
	if {$::sd_replaceProc!=0} \
	{
		lappend result "-replacescript";
	}
	return $result
}

# Create a limited set of search options for the "find/replace next" and "find selection" functions.
# these don't include search direction (since that's part of the command), and they
# don't include "limit scope", since you generally don't want that with those commands
# either.
proc CreateLimitedSearchOptions {window} \
{
	set result ""

	# If the window is in UTF8 mode, then search in UTF8 mode
	if {[getutf8 $window]} \
	{
		lappend result "-utf8";
	}

	# Append options from search dialog
	if {$::sd_wrapAround!=0} \
	{
		lappend result "-wrap";
	}
	if {$::sd_ignoreCase!=0} \
	{
		lappend result "-ignorecase";
	}
	if {$::sd_selectionExpression!=0} \
	{
		lappend result "-regex";
	}
	if {$::sd_useSubexpression!=0} \
	{
		lappend result "-usesubexpression";
		lappend result $::sd_subexpressionIndex;
	}
	if {$::sd_replaceProc!=0} \
	{
		lappend result "-replacescript";
	}
	return $result
}

# Ask the user just what he wishes to search for, and how, then do it in all open windows if he does not cancel.
proc AskSearchAll {} \
{
	set searchType [searchdialog "" findBuffer replaceBuffer ::sd_backwards ::sd_wrapAround ::sd_ignoreCase ::sd_limitScope ::sd_selectionExpression ::sd_useSubexpression ::sd_subexpressionIndex ::sd_replaceProc]
	foreach window [windowlist] \
	{
		switch $searchType \
		{
			"find"
			{
				eval [list FindBeep $window findBuffer] [CreateSearchOptions $window]
			}
			"findall"
			{
				eval [list FindAllBeep $window findBuffer] [CreateSearchOptions $window]
			}
			"replace"
			{
				eval [list ReplaceBeep $window findBuffer replaceBuffer] [CreateSearchOptions $window]
			}
			"replaceall"
			{
				eval [list ReplaceAllBeep $window findBuffer replaceBuffer] [CreateSearchOptions $window]
			}
		}
	}
	okdialog "All windows completed"
}

# Ask the user just what he wishes to search for, and how, then do it if he does not cancel.
proc AskSearch {window} \
{
	set searchType [searchdialog "" findBuffer replaceBuffer ::sd_backwards ::sd_wrapAround ::sd_ignoreCase ::sd_limitScope ::sd_selectionExpression ::sd_useSubexpression ::sd_subexpressionIndex ::sd_replaceProc]
	switch $searchType \
	{
		"find"
		{
			eval [list FindBeep $window findBuffer] [CreateSearchOptions $window]
		}
		"findall"
		{
			eval [list FindAllBeep $window findBuffer] [CreateSearchOptions $window]
		}
		"replace"
		{
			eval [list ReplaceBeep $window findBuffer replaceBuffer] [CreateSearchOptions $window]
		}
		"replaceall"
		{
			eval [list ReplaceAllBeep $window findBuffer replaceBuffer] [CreateSearchOptions $window]
		}
	}
}

# Used for find same backwards/forwards.
proc FindNext {window backward} \
{
	set flags [CreateLimitedSearchOptions $window]
	# if asked to search backwards, then add this flag to the list
	if {$backward!=0} \
	{
		lappend flags "-backward"
	} \
	else \
	{
		lappend flags "-forward"
	}
	eval [list FindBeep $window findBuffer] $flags
}

# if selection then first copy the selection and then do FindAllBeep
# if no selection then try to find what is in the findBuffer
proc FindSelectionNext {window backward} \
{
	set ends [getselectionends $window]
	set start [lindex $ends 0]
	set end [lindex $ends 1]
	if {$start!=$end} \
	{
		copy $window findBuffer
	}
	FindNext $window $backward
}

# Used for replace same backwards/forwards.
proc ReplaceNext {window backward} \
{
	set flags [CreateLimitedSearchOptions $window]
	# if asked to search backwards, then add this flag to the list
	if {$backward!=0} \
	{
		lappend flags "-backward"
	} \
	else \
	{
		lappend flags "-forward"
	}
	eval [list ReplaceBeep $window findBuffer replaceBuffer] $flags
}

# Replace all selections with some given text.
proc ReplaceSelections {window} \
{
	TextToBuffer tempFindBuffer {([^]+)};							# load up the expression (find anything, mark as \0)
	TextToBuffer tempReplaceBuffer [set ::replaceSelectionsChoice [textdialog "Replacement string?\nNOTE: use '\\0' to refer to current selection." $::replaceSelectionsChoice]];	# load up the replacement
	ReplaceAllBeep $window tempFindBuffer tempReplaceBuffer -regex -limitscope;	# do the replacement
}

# Shift the text but attempt to leave the selection as little disturbed as possible
proc HandleShifts {buffer} \
{
	set numSelections [lindex [selectioninfo $buffer] 6]

	if {$numSelections>0} \
	{
		# if one selection, try not to "columnarize" it
		if {$numSelections==1} \
		{
			catch {closebuffer tempBuffer};						# get rid of any lingering tempBuffer
			newbuffer tempBuffer;								# create a place to do some messing around in
			copy $buffer tempBuffer;							# move text in question to temp buffer
			setselectionends tempBuffer 0 0;					# move to top for replace

			set start [lindex [getselectionends $buffer] 0];	# get start of selection

			if {[catch {replaceall tempBuffer tempFindBuffer tempReplaceBuffer -regex} message]==0} \
			{
				if {[llength $message]} \
				{
					setselectionends tempBuffer 0 0;			# get rid of selections in temp buffer so paste does what we want
					paste $buffer tempBuffer;					# place results back, write over old selection
					set length [lindex [textinfo tempBuffer] 1]
					setselectionends $buffer $start [expr $start+$length] ;	# re-select the block
				} \
				else \
				{
					beep
				}
			} \
			else \
			{
				okdialog $message
			}
			closebuffer tempBuffer;
		} \
		else \
		{
			if {[catch {replaceall $buffer tempFindBuffer tempReplaceBuffer -regex -limitscope} message]==0} \
			{
				if {[llength $message]==0} {beep}
			} \
			else \
			{
				okdialog $message
			}
		}
	} \
	else \
	{
		beep
	}
}

# Request a Unicode code point, format as UTF-8, and insert at the current location.
# NOTE: Tcl handles UTF-8 internally, and we _could_ just send the codepoint to
# the "format" command to get it converted, but Tcl only handles 16 bits of codepoint
# and we want to handle more than that.
proc InsertUnicode {buffer} \
{
	set string ""
	set codePoint [expr [textdialog "Enter Unicode Codepoint Value:" ""]]
	if {$codePoint>=0} \
	{
		if {$codePoint<(1<<7)} \
		{
			append string [format %c $codePoint]
		} \
		elseif {$codePoint<(1<<(5+6))} \
		{
			append string [format %c [expr 0xC0+($codePoint>>6)]]
			append string [format %c [expr 0x80+($codePoint&0x3F)]]
		} \
		elseif {$codePoint<(1<<(4+6+6))} \
		{
			append string [format %c [expr 0xE0+($codePoint>>12)]]
			append string [format %c [expr 0x80+(($codePoint>>6)&0x3F)]]
			append string [format %c [expr 0x80+($codePoint&0x3F)]]
		} \
		elseif {$codePoint<(1<<(3+6+6+6))} \
		{
			append string [format %c [expr 0xF0+($codePoint>>18)]]
			append string [format %c [expr 0x80+(($codePoint>>12)&0x3F)]]
			append string [format %c [expr 0x80+(($codePoint>>6)&0x3F)]]
			append string [format %c [expr 0x80+($codePoint&0x3F)]]
		} \
		else \
		{
			beep
			okdialog "Value too large"
			return
		}
		insertbinary $buffer $string
	} \
	else \
	{
		beep
		okdialog "Value must be positive"
		return
	}
}

# Align text left, removing all tabs or spaces at the start of lines in the selection.
proc AlignLeft {buffer} \
{
	TextToBuffer tempFindBuffer {^[\t ]+(.*)|(.+)};				# load up the expression
	TextToBuffer tempReplaceBuffer {\0\1};						# load up the replacement
	HandleShifts $buffer;
}

# Shift text left, removing tabs or spaces at the start of lines in the selection.
proc ShiftLeft {buffer} \
{
	TextToBuffer tempFindBuffer {^[\t ](.*)|(.+)};				# load up the expression (expression is slightly strange, so that lines which are not altered are not deselected)
	TextToBuffer tempReplaceBuffer {\0\1};						# load up the replacement
	HandleShifts $buffer;
}

# Shift text right, adding tabs at the start of lines in the selection.
proc ShiftRight {buffer} \
{
	TextToBuffer tempFindBuffer {^(.+)};						# load up the expression
	TextToBuffer tempReplaceBuffer {	\0};					# load up the replacement
	HandleShifts $buffer;
}

# Search forward from mark to the first closing brace which isn't matched
# by an opening brace.
proc LocateUnmatchedClosingBrace {buffer mark} \
{
	set depth 0
	set done 0
	TextToBuffer tempFindBuffer {\{|\}};						# find either an opening or closing brace
	while {!$done} \
	{
		if {[catch {find $buffer tempFindBuffer -regex -sourcemark $mark -destmark $mark} message]==0} \
		{
			if {[llength $message]!=0} \
			{
				if {[lindex [selectedtextlist $buffer -mark $mark] 0]=="\{"} \
				{
					set depth [expr $depth+1];					# go deeper
				} \
				else \
				{
					if {$depth!=0} \
					{
						set depth [expr $depth-1];				# back up
					} \
					else \
					{
						return $message;						# return the location of the closing brace
					}
				}
			} \
			else \
			{
				set done 1;
			}
		} \
		else \
		{
			return -code error $message
		}
	}
	return;
}

# Search backward from mark to the first opening brace which isn't matched
# by a closing brace.
proc LocateUnmatchedOpeningBrace {buffer mark} \
{
	set depth 0
	set done 0
	TextToBuffer tempFindBuffer {\{|\}};						# find either an opening or closing brace
	while {!$done} \
	{
		if {[catch {find $buffer tempFindBuffer -backward -regex -sourcemark $mark -destmark $mark} message]==0} \
		{
			if {[llength $message]!=0} \
			{
				if {[lindex [selectedtextlist $buffer -mark $mark] 0]=="\}"} \
				{
					set depth [expr $depth+1];					# go deeper
				} \
				else \
				{
					if {$depth!=0} \
					{
						set depth [expr $depth-1];				# back up
					} \
					else \
					{
						return $message;						# return the location of the opening brace
					}
				}
			} \
			else \
			{
				set done 1;
			}
		} \
		else \
		{
			return -code error $message
		}
	}
	return;
}

# Search backward from the current position to the first open brace '{'.
# then search forward to find the matching close brace '}'.
# Select all the lines between the lines the two braces are on.
proc SelectLinesBetweenBraces {buffer} \
{
	openmark $buffer _findTemp;									# open the destination mark for the search
	copyselection $buffer -destmark _findTemp;					# copy the mark over the buffer's selection
	if {[catch {LocateUnmatchedOpeningBrace $buffer _findTemp} message]==0} \
	{
		# make sure something was found
		if {[llength $message]!=0} \
		{
			if {[catch {LocateUnmatchedClosingBrace $buffer _findTemp} message2]==0} \
			{
				if {[llength $message2]!=0} \
				{
					set startPosition [lindex $message 1];		# get position of the end of the top match
					set endPosition [lindex $message2 0];		# get position of the start of the bottom match
					set topLine [lindex [positiontolineoffset $buffer $startPosition] 0];		# get line where the top brace was
					set bottomLine [lindex [positiontolineoffset $buffer $endPosition] 0];		# get line where the bottom brace was

					set topLine [expr $topLine+1];				# move in
					if {$topLine<$bottomLine} \
					{
						setselectionends $buffer [lineoffsettoposition $buffer $topLine 0] [lineoffsettoposition $buffer $bottomLine 0];	# select result
					} \
					else \
					{
						setselectionends $buffer $startPosition $endPosition;	# just select between the braces
					}
					HomeWindowToSelectionStart $buffer
				} \
				else \
				{
					beep;										# couldn't find matching close brace
				}
			} \
			else \
			{
				okdialog $message2;								# something failed, report it
			}
		} \
		else \
		{
			beep;												# couldn't find an open brace
		}
	} \
	else \
	{
		okdialog $message;										# something failed, report it
	}
	closemark $buffer _findTemp
}

# Unselect any whitspace which is currently selected.
# This is sometimes useful during columnar select when tabs or spaces
# get in the way
proc UnselectWhitespace {buffer} \
{
	TextToBuffer tempFindBuffer {[^ \t\n]+};					# load up the expression to match any non-white characters
	if {[catch {findall $buffer tempFindBuffer -regex -limitscope} message]==0} \
	{
		if {[llength $message]==0} \
		{
			# no non-white characters were located, so selection is ALL white, just eliminate the selection
			set start [lindex [getselectionends $buffer] 0];	# get start of selection
			setselectionends $buffer $start $start;				# reduce selection to nothing
		}
	} \
	else \
	{
		okdialog $message
	}
}

# Unselect any characters which don't appear to look like
# identifiers in 'C'.
proc UnselectNonIdentifier {buffer} \
{
	TextToBuffer tempFindBuffer {[_A-Za-z0-9]+};				# load up the expression to match only identifier characters
	if {[catch {findall $buffer tempFindBuffer -regex -limitscope} message]==0} \
	{
		if {[llength $message]==0} \
		{
			# no identifier characters were located, just eliminate the selection
			set start [lindex [getselectionends $buffer] 0];	# get start of selection
			setselectionends $buffer $start $start;				# reduce selection to nothing
		}
	} \
	else \
	{
		okdialog $message
	}
}

# Ask the user for the command to pipe through.
proc PipeCommandGet {question} \
{
	set ::pipeSelectionsChoice [textdialog $question $::pipeSelectionsChoice];
}

# Pipe each selection of a window through a shell command, collect output, and replace selections.
proc PipeSelection {window command} \
{
	TextToBuffer tempFindBuffer {[^]+};							# load up the expression (find absolutely anything)
	TextToBuffer tempReplaceBuffer "catch \{exec -keepnewline $command <<\$found\} message; set message";	# load up the replacement, return results of command
	ReplaceAllBeep $window tempFindBuffer tempReplaceBuffer -regex -limitscope -replacescript;			# do the replacement
}

proc PipeWindowCommandGet {question} \
{
	set ::pipeWindowsChoice [textdialog $question $::pipeWindowsChoice];
}

# Pipe the entire contents of a given window through a shell command.
proc PipeWindow {window command} \
{
	TextToBuffer tempFindBuffer {[^]+};							# load up the expression (find absolutely anything)
	TextToBuffer tempReplaceBuffer "catch \{exec -keepnewline $command <<\$found\} message; set message";	# load up the replacement, return results of command
	setselectionends $window 0 0;								# move to top for replace
	replace $window tempFindBuffer tempReplaceBuffer -regex -replacescript;			# do the replacement
}

# Convert all letters of the given selection to upper case.
proc UpperCaseSelection window \
{
	TextToBuffer tempFindBuffer {[^]+};							# load up the expression (find anything)
	TextToBuffer tempReplaceBuffer "string toupper \$found";	# load up the replacement
	ReplaceAllBeep $window tempFindBuffer tempReplaceBuffer -regex -limitscope -replacescript;	# do the replacement
}

# Convert all letters of the given selection to lower case.
proc LowerCaseSelection window \
{
	TextToBuffer tempFindBuffer {[^]+};							# load up the expression (find anything)
	TextToBuffer tempReplaceBuffer "string tolower \$found";	# load up the replacement
	ReplaceAllBeep $window tempFindBuffer tempReplaceBuffer -regex -limitscope -replacescript;	# do the replacement
}

# Locate all the numbers in the current selection, and increment them by the given amount.
proc IncrementSelection {window amount} \
{
	TextToBuffer tempFindBuffer {-?(([0-9]+(\.[0-9]+)?)|(\.[0-9]+))};		# load up the expression (find any (possibly floating point) number including negative ones)
	TextToBuffer tempReplaceBuffer "expr [string trimleft \$found 0] + $amount";	# load up the replacement, return the result of adding amount
	ReplaceAllBeep $window tempFindBuffer tempReplaceBuffer -regex -limitscope -replacescript;	# do the replacement
}

# Replace each selection by an incrementing number.
proc EnumerateSelection {window startAt amount} \
{
	set ::enumStart $startAt
	TextToBuffer tempFindBuffer {.+};							# load up the expression (find anything)
	TextToBuffer tempReplaceBuffer "set temp \$::enumStart;incr ::enumStart $amount;format %d \$temp";		# load up the replacement, return an incremented number
	ReplaceAllBeep $window tempFindBuffer tempReplaceBuffer -regex -limitscope -replacescript;	# do the replacement
}

# Report the sum of all selected numbers.
proc SumSelection {window} \
{
	SetMark $window temp;										# remember what was selected, we will not disturb it
	TextToBuffer tempFindBuffer {-?(([0-9]+(\.[0-9]+)?)|(\.[0-9]+))};		# load up the expression (find any (possibly floating point) number including negative ones)
	if {[catch {findall $window tempFindBuffer -regex -limitscope} message]==0} \
	{
		if {[llength $message]} \
		{
			set total 0
			foreach number [selectedtextlist $window] \
			{
				set total [expr [string trimleft $number 0]+$total];	# convert number with leading 0's to DECIMAL, not octal!
			}
			okdialog "Sum = $total"
		} \
		else \
		{
			beep
		}
	} \
	else \
	{
		okdialog $message
	}
	GotoMark $window temp;									# put back the user's selection
	closemark $window temp;									# get rid of temp selection
}

# Sort the selections, and replace them in sorted order.
proc SortSelection {window} \
{
	catch {unset ::sortedArray}
	set ::sortedIndex 0
	foreach element [lsort [selectedtextlist $window]] \
	{
		set ::sortedArray($::sortedIndex) $element;				# copy elements into an array to speed things up
		incr ::sortedIndex
	}
	set ::sortedIndex 0
	TextToBuffer tempFindBuffer {[^]+};							# load up the expression (find absolutely anything)
	TextToBuffer tempReplaceBuffer {incr ::sortedIndex;set ::sortedArray([expr $::sortedIndex-1])};		# load up the replacement, return an entry from the array
	ReplaceAllBeep $window tempFindBuffer tempReplaceBuffer -regex -limitscope -replacescript;	# do the replacement
	catch {unset ::sortedArray}
	unset ::sortedIndex
}

# Replace the selections in reverse order.
proc ReverseSelection {window} \
{
	catch {unset ::reverseArray}
	set ::reverseIndex 0
	foreach element [selectedtextlist $window] \
	{
		set ::reverseArray($::reverseIndex) $element;			# copy elements into an array to speed things up
		incr ::reverseIndex
	}
	TextToBuffer tempFindBuffer {[^]+};							# load up the expression (find absolutely anything)
	TextToBuffer tempReplaceBuffer {incr ::reverseIndex -1;set ::reverseArray($::reverseIndex)};	# load up the replacement, return an entry from the array
	ReplaceAllBeep $window tempFindBuffer tempReplaceBuffer -regex -limitscope -replacescript;	# do the replacement
	catch {unset ::reverseArray}
	unset ::reverseIndex
}

# Get a manual page, and dump it into a window.
proc ManPage params \
{
	set name [newbuffer [NewTitledBufferName "Man Page '$params'"]];			# make a buffer to hold the man page
	# move stdin to /dev/null so that man won't look at current terminal width
	execute $name "catch \{exec man -Tascii $params </dev/null\} result; set result";

# man thinks it is outputing to a printer where it can backspace over stuff to get effects
# like BOLD, or dot points. The replace strips out the mess

# first, go through and make keywords out of anything man was trying to make bold
	setselectionends $name 0 0;						# move to the top of the man page buffer
	TextToBuffer tempFindBuffer {((.)\x08\1)+};		# load up the expression
	catch {findall $name tempFindBuffer -regex};	# do this, ignore any errors
	selectiontostyle $name $::style_keyword;		# set what we found to ::style_keyword

# now, go through and make strings out of anything man was trying to make underlines
	setselectionends $name 0 0;						# move to the top of the man page buffer
	TextToBuffer tempFindBuffer {(_\x08[^\n_])+};	# load up the expression
	catch {findall $name tempFindBuffer -regex};	# do this, ignore any errors
	selectiontostyle $name $::style_string;			# set what we found to ::style_string

# remove all the backspace characters
	setselectionends $name 0 0;						# move to the top of the man page buffer
	TextToBuffer tempFindBuffer {.\x08};			# load up the expression
	TextToBuffer tempReplaceBuffer {};				# replace with nothing
	catch {replaceall $name tempFindBuffer tempReplaceBuffer -regex};	# do this, ignore any errors

# replace special double-quote chars
	setselectionends $name 0 0;										 	# move to the top of the man page buffer
	TextToBuffer tempFindBuffer {(\xE2\x80[\x98,\x99]){2}};				# load up the expression
	TextToBuffer tempReplaceBuffer {"};									# replace with nothing
	catch {replaceall $name tempFindBuffer tempReplaceBuffer -regex};	# do this, ignore any errors

# replace special single-quote chars
	setselectionends $name 0 0;											# move to the top of the man page buffer
	TextToBuffer tempFindBuffer {\xE2\x80[\x98,\x99]};					# load up the expression
	TextToBuffer tempReplaceBuffer {'};									# replace with nothing
	catch {replaceall $name tempFindBuffer tempReplaceBuffer -regex};	# do this, ignore any errors

# get rid of any huge page gaps
	setselectionends $name 0 0;						# move to the top of the buffer again
	TextToBuffer tempFindBuffer {^\n{2,}};			# load up the expression
	TextToBuffer tempReplaceBuffer "\n";			# replace with nothing
	catch {replaceall $name tempFindBuffer tempReplaceBuffer -regex};	# do this, ignore any errors
	setselectionends $name 0 0;						# move to the top of the buffer again

# highlight any links that might be present
	setselectionends $name 0 0;						# move to the top of the buffer again
	TextToBuffer tempFindBuffer {<[A-Za-z][0-9A-Za-z+.\-]*://[0-9A-Za-z.\-_/~@?=+&#:%]+};	# load up the expression
	catch {findall $name tempFindBuffer -regex};	# go find them
	selectiontostyle $name $::style_link;			# set what we found to ::style_link

	setselectionends $name 0 0;						# move to the top of the buffer one final time

	flushundos $name;								# get rid of any undos we made here, just to be nice
	cleardirty $name;								# make it non-modified

	OpenWindow $name $::staggerOpenX $::staggerOpenY [expr 8*80] $::windowHeight 1 8 $::defaultFontName "Man Page" ""
	UpdateStaggerPosition;
}

# Refill a selected paragraph(s), so that the lines of text do not extend past the given
# column. This is a little ugly, but it does the job quite nicely. This also works if the
# selection is columnar (can be used to refill comments)
proc RefillSelection buffer \
{
	set limit [expr [set ::refillLimitChoice [textdialog "Max length:" $::refillLimitChoice]]];		# get length limit from user

	set ends [getselectionends $buffer]
	set start [lindex $ends 0]
	set end [lindex $ends 1]
	if {$start!=$end} \
	{
		catch {closebuffer tempBuffer};					# get rid of any lingering tempBuffer
		newbuffer tempBuffer;							# create a place to do some messing around in
		copy $buffer tempBuffer;						# move text in question to temp buffer

		TextToBuffer tempFindBuffer "(^\[ \\t\]+)|(\[ \t\]+$)";		# remove white space at starts and ends of lines
		TextToBuffer tempReplaceBuffer {};
		setselectionends tempBuffer 0 0;				# move to top for replace
		catch {replaceall tempBuffer tempFindBuffer tempReplaceBuffer -regex};	# do this, ignore any errors

		TextToBuffer tempFindBuffer "\[ \\t\]+";		# make tabs, or multiple spaces into single spaces
		TextToBuffer tempReplaceBuffer { };
		setselectionends tempBuffer 0 0;				# move to top for replace
		catch {replaceall tempBuffer tempFindBuffer tempReplaceBuffer -regex};	# do this, ignore any errors

		TextToBuffer tempFindBuffer "(^\\n)|(\\n)";		# make all newlines at the ends of lines into spaces
		TextToBuffer tempReplaceBuffer {if {[string length $1] > 0} {set found { }}; set found};
		selectall tempBuffer;							# select all so replace will start at the bottom (replacing backwards)
		catch {replaceall tempBuffer tempFindBuffer tempReplaceBuffer -backward -regex -replacescript};	# do this, ignore any errors

		TextToBuffer tempFindBuffer "(.\{1,$limit\})((\[ \\t\]+)|(.$))";	# cut lines to length at whitespace
		TextToBuffer tempReplaceBuffer {if {[string length $2] > 0} {format "%s\n" $0} else {format "%s%s" $0 $3}};
		setselectionends tempBuffer 0 0;				# move to top for replace
		catch {replaceall tempBuffer tempFindBuffer tempReplaceBuffer -regex -replacescript};	# do this, ignore any errors

		setselectionends tempBuffer 0 0;				# move to top of temp buffer

		set startNumSelections [lindex [selectioninfo $buffer] 6];
		if {$startNumSelections>1} \
		{
			TextToBuffer tempFindBuffer "^.*$";			# select all the lines individually
			catch {findall tempBuffer tempFindBuffer -regex};	# do this, ignore any errors

			set endNumSelections [lindex [selectioninfo tempBuffer] 6];
			if {$endNumSelections>$startNumSelections} \
			{
				addselectionendlist tempBuffer [list [list [lineoffsettoposition tempBuffer $startNumSelections 0] [lineoffsettoposition tempBuffer [expr $endNumSelections+1] 0]]]
			}
		}

		paste $buffer tempBuffer;					# place results back, write over old selection(s)

		closebuffer tempBuffer;
	} \
	else \
	{
		beep;
	}
}

# Check spelling in buffer, make a new window containing the misspelled words.
proc SpellDocument buffer \
{
	set name [newbuffer [NewTitledBufferName Spell]];		# make a buffer to hold the output of the spell command
	SetMark $buffer temp;							# remember what was selected, we will not disturb it
	selectall $buffer;								# make a selection of all of the text in the passed buffer
	execute $name {catch {exec aspell list << [lindex [selectedtextlist $buffer] 0] | sort -u} result; set result};
	GotoMark $buffer temp;							# put selection back
	setselectionends $name 0 0;						# move to the top of the output buffer

	flushundos $name;								# get rid of any undos we made here, just to be nice
	cleardirty $name;								# make it non-modified
	openwindow $name $::staggerOpenX $::staggerOpenY [expr 7*50] $::windowHeight 1 8 $::defaultFontName "Simple White";	# make a new window, set width to roughly 50 columns of an 7 pixel wide font

	UpdateStaggerPosition;
	closemark $buffer temp;							# get rid of temp selection
}

# Get an address, and mail the contents of buffer
# to that address.
# NOTE: this would be much better handled with a Tk dialog
proc MailDocument buffer \
{
	# get recipient
	set recipient [set ::mailRecipientChoice [textdialog "Mail $buffer To:" $::mailRecipientChoice]]
	# get subject
	set subject [set ::mailSubjectChoice [textdialog "Subject:" $::mailSubjectChoice]]

	catch {closebuffer tempBuffer};					# get rid of any lingering tempBuffer
	newbuffer tempBuffer;							# create temp buffer
	SetMark $buffer temp;							# remember what was selected, we will not disturb it
	selectall $buffer;								# make a selection of all of the text in the passed buffer
	copy $buffer tempBuffer;						# copy selection to temp buffer
	GotoMark $buffer temp;							# put selection back
	closemark $buffer temp;							# get rid of temp selection
	setselectionends tempBuffer 0 0;				# move to the top of the temp buffer
	insert tempBuffer "To: $recipient\nSubject: $subject\n\n";	# insert recipient and subject to buffer
	set result 0;									# predefine no error happend
	selectall tempBuffer;							# select what to mail

	if {[catch {exec /usr/lib/sendmail -t << [lindex [selectedtextlist tempBuffer] 0]} result]!=0} \
	{
		okdialog "Mail status:\n\n$result\n"
	}
	closebuffer tempBuffer;								# get rid of temp buffer
}

# mail contents of chosen windows to chosen recipients
proc MailWindows {} \
{
	foreach buffer [listdialog "Choose Window(s) to Mail:" [windowlist]] \
	{
		MailDocument $buffer
	}
}

# Strip the white space at the ends of lines, attempt to leave the selection as it was.
# return the number of characters removed
proc StripWhite {window} \
{
	set initialChars [lindex [textinfo $window] 1];			# find out how many characters we have at the start
	set numRemoved 0;
	SetMark $window temp;									# remember what was selected, we will not disturb it
	TextToBuffer tempFindBuffer {[ \t]+$};					# load up the expression
	TextToBuffer tempReplaceBuffer {};						# replace with nothing
	setselectionends $window 0 0;							# move to the top of the buffer to perform the search
	if {[catch {replaceall $window tempFindBuffer tempReplaceBuffer -regex} message]==0} \
	{
		set finalChars [lindex [textinfo $window] 1];		# find out how many characters we have now that we are done
		set numRemoved [expr $initialChars-$finalChars];
	} \
	else \
	{
		okdialog $message
	}
	GotoMark $window temp;									# put back the user's selection
	closemark $window temp;									# get rid of temp selection
	return $numRemoved
}

# Set line termination to the given sequence
proc SetLineTermination {window termination} \
{
	SetMark $window temp;									# remember what was selected, we will not disturb it
	TextToBuffer tempFindBuffer {\r\n|\r|\n};				# load up the expression
	TextToBuffer tempReplaceBuffer $termination;			# replace with given characters
	setselectionends $window 0 0;							# move to the top of the buffer to perform the search
	catch {replaceall $window tempFindBuffer tempReplaceBuffer -regex}
	GotoMark $window temp;									# put back the user's selection
	closemark $window temp;									# get rid of temp selection
}

# insert a character into window, and then home the window so that
# the character is visible
proc InsertAndHome {window character} \
{
	insert $window $character;								# insert the character
	HomeWindowToSelectionStart $window -lenient;			# make sure it is visible
}

# Inserts newline if the cursor position is above a given value
# This is a simple way to implement wrapping lines at word boundaries
# Thanks to Juergen Reiss
proc WrapLine buffer \
{
	set ends [getselectionends $buffer];					# get end of selection
	set start [lindex $ends 0];								# line number
	set pos [lindex [positiontolineoffset $buffer $start] 1];# get column number
	if {$pos>$::lineWrapColumn} \
	{
		InsertAndHome $buffer "\n";							# insert newline
	} \
	else \
	{
		InsertAndHome $buffer " ";							# insert space
	}
}

# Changes the status of line-wrapping
proc WrapOnOff {} \
{
	set temp "off";
	if {$::lineWrap} \
	{
		set temp "on";
	}
	set ::lineWrap [yesnodialog "Line wrapping is currently '$temp'\nDo you want to have line wrapping on?"];
	if {$::lineWrap} \
	{
		bindkey {space}	{WrapLine [ActiveWindowOrBeep]};		# when space is hit, see if above the given column and insert newline if needed
	} \
	else \
	{
		unbindkey {space}
	}
}

# Changes the column at which lines wrap
proc WrapNewColumn {} \
{
	set ::lineWrapColumn [expr [textdialog "New line wrap column:" $::lineWrapColumn]];		# get new column limit from user
}

# Get text to evaluate from window, then stick in a newline.
proc EvalText {window} \
{
	set ends [getselectionends $window]
	set start [lindex $ends 0]
	set end [lindex $ends 1]
	if {$start==$end} \
	{
		set line [lindex [positiontolineoffset $window $start] 0];			# get line we want to select
		set start [lineoffsettoposition $window $line 0];					# point to the start
		set end [lineoffsettoposition $window [expr $line+1] 0];			# point to the end
	}
	if {[lindex [positiontolineoffset $window $end] 1]==0} \
	{
		if {$start!=$end} \
		{
			incr end -1;													# move back past last newline
		}
	}
	setselectionends $window $start $end;									# make only one selection, even if more than one
	set result [lindex [selectedtextlist $window] 0]\n;						# get contents of selection to return (add in new line)
	setselectionends $window $end $end;										# move to the end
	breakundo $window;														# force a break in the undo stream for this window
	InsertAndHome $window "\n";												# add newline
	set result;																# return this
}

# When a shell command arrives, this handles it
# NOTE: ShellCommand is a procedure that the editor expects to be defined.
# The editor calls this procedure in response to certain events.
# The following is a list of the commands that are currently defined:
# initialargs	-- sent when the editor begins, args is the list of command line arguments
# open			-- editor is asked to open documents, args is path name list
# close			-- editor is asked to close documents, args is buffer list
# quit			-- editor is asked to quit, args is not used
proc ShellCommand {command args} \
{
	switch $command \
	{
		"initialargs"
		{
			OpenList $args;							# later, if we want, we can interpret this for command line switches, or whatever
		}
		"open"
		{
			OpenList $args
		}
		"close"
		{
			foreach window $args \
			{
				AskClose $window
			}
		}
		"quit"
		{
			TryToQuit
		}
	}
}

# display the e93 about box
proc AboutBox {} \
{
	okdialog "[version]\nTcl: [info tclversion]\nStartup Script: $::SCRIPTPATH\n\n$::tcl_platform(os) $::tcl_platform(osVersion) $::tcl_platform(machine)\nProcess ID: [pid]\n\nNumber of open windows: [llength [windowlist]]\nNumber of open buffers: [llength [bufferlist]]\n\nFor the latest version, visit www.e93.org"
}

# Add another directory path to the directory menu.
proc AddDirectoryMenu {path} \
{
	addmenu {Directory} LASTCHILD 1 "$path" {} "cd \{$path\}"
}

# Attempt to process the user preferences from the user's home directory.
# If they cannot be found, attempt to process from the system directory.
proc ProcessUserPreferences {} \
{
	set defaultPrefs [file join $::sysFilesDir $::prefsFile];		# point to system preferences file
	set userPrefs [file join $::userFilesDir $::prefsFile];	# point to user's preferences file

	set file $defaultPrefs;											# use system preferences file if user doesn't have their own
	if {[file exists $userPrefs]} \
	{
		set file [glob $userPrefs];									# use user's file
	}

	if {[catch {source $file}]!=0} \
	{
		okdialog "Error processing preferences:\n\n$::errorInfo";	# if the file contains errors, don't let them kill us, just report and carry on
	}
}

# This function opens the user's preferences file for editing.
# When the user saves the file, it is executed (to adjust the preferences
# immediately.
proc AdjustUserPreferences {} \
{
	set defaultPrefs [file join $::sysFilesDir $::prefsFile];		# point to system preferences file
	set userPrefs [file join $::userFilesDir $::prefsFile];			# point to user's preferences file

	if {![file exists $userPrefs]} \
	{
		file mkdir $::userFilesDir;									# make sure the directory exists
		file copy $defaultPrefs $userPrefs;							# start out user with a copy of the system preferences
	}

	set file [glob $userPrefs];
	if {[catch {OpenDefaultWindowRecordMtime $file} errorMessage]==0} \
	{
		setbuffervariable $file aftersave "ProcessUserPreferences";	# after file is saved, reload the preferences
	} \
	else \
	{
		okdialog "Error loading preferences:\n\n$errorMessage";
	}
}

# Start by processing user preferences
ProcessUserPreferences

# Define the menus.
addmenu {} LASTCHILD 1 "e93" "" ""
	addmenu {e93} LASTCHILD 1 "About..."						{}				{AboutBox}
	addmenu {e93} LASTCHILD 0 "space0"							{\S}			{}

	addmenu {e93} LASTCHILD 1 "Help for e93"					{}				{OpenList [file join $::sysFilesDir $::docsDir README.e93]}
	addmenu {e93} LASTCHILD 1 "Help for regular expressions"	{}				{OpenList [file join $::sysFilesDir $::docsDir README.regex]}
	addmenu {e93} LASTCHILD 1 "Help for syntax maps"			{}				{OpenList [file join $::sysFilesDir $::docsDir README.syntaxmaps]}
	addmenu {e93} LASTCHILD 0 "space2"							{\S}			{}
	addmenu {e93} LASTCHILD 1 "Preferences"						{}				{AdjustUserPreferences}
	addmenu {e93} LASTCHILD 0 "space3"							{\S}			{}
	addmenu {e93} LASTCHILD 1 "Themes"							{}				{}

	# get all themes
	foreach theme [lsort [ThemeList]] \
	{
		addmenu {e93 Themes} LASTCHILD 1 $theme					{}				"ThemeSet $theme"
	}

addmenu {} LASTCHILD 0 "space0" {\S}	{}

addmenu {} LASTCHILD 1 "File" "" ""
	addmenu {File} LASTCHILD 1 "New"							{\KCmd+n}				{NewUntitledWindow}
	addmenu {File} LASTCHILD 1 "Open..."						{\KCmd+o}				{OpenList [opendialog "Open File:"]}
	addmenu {File} LASTCHILD 1 "Open Selection"					{\KCmd+d}				{SmartOpen [ActiveWindowOrBeep]}
	addmenu {File} LASTCHILD 1 "Include..."						{}						{IncludeList [set window [ActiveWindowOrBeep]] [opendialog "Include:"];HomeWindowToSelectionStart $window}
	addmenu {File} LASTCHILD 0 "space0"							{\S}					{}
	addmenu {File} LASTCHILD 1 "Close"							{\KCmd+w}				{AskClose [ActiveWindowOrBeep]}
	addmenu {File} LASTCHILD 1 "Close All"						{\KCmd+Shift+w}			{foreach window [windowlist] {AskClose $window}}
	addmenu {File} LASTCHILD 1 "Save"							{\KCmd+s}				{AskSave [ActiveWindowOrBeep]}
	addmenu {File} LASTCHILD 1 "Save As..."						{\KCmd+Shift+s}			{AskSaveAs [ActiveWindowOrBeep]}
	addmenu {File} LASTCHILD 1 "Save To..."						{}						{AskSaveTo [ActiveWindowOrBeep]}
	addmenu {File} LASTCHILD 1 "Save All"						{}						{AskSaveAll}
	addmenu {File} LASTCHILD 1 "Revert To Saved"				{}						{AskRevert [ActiveWindowOrBeep]}
	addmenu {File} LASTCHILD 0 "space1"							{\S}					{}
	addmenu {File} LASTCHILD 1 "Print..."						{\KCmd+p}				{PrintBuffer [ActiveWindowOrBeep]}
	addmenu {File} LASTCHILD 1 "Print Selection..."				{\KCmd+Shift+p}			{PrintSelection [ActiveWindowOrBeep]}
	addmenu {File} LASTCHILD 1 "Print Windows..."				{}						{PrintWindows}
	addmenu {File} LASTCHILD 0 "space2"							{\S}					{}
	addmenu {File} LASTCHILD 1 "Quit"							{\KCmd+q}				{TryToQuit}

addmenu {} LASTCHILD 1 "Directory" "" ""
	addmenu {Directory} LASTCHILD 1 "Show Current Directory"	{}						{okdialog "Current directory:\n\n[pwd]"}
	addmenu {Directory} LASTCHILD 1 "Set Directory..."			{}						{set path [pathdialog "Choose directory:"]; cd "$path"; AddDirectoryMenu "$path"}
	addmenu {Directory} LASTCHILD 1 "space0"					{\S}					{}

addmenu {} LASTCHILD 1 "Edit" "" ""
	addmenu {Edit} LASTCHILD 1 "Undo/Redo Toggle"				{\KCmd+z}				{if {[undotoggle [set window [ActiveWindowOrBeep]]]!=0} {HomeWindowToSelectionStart $window} else {beep}}
	addmenu {Edit} LASTCHILD 1 "Undo"							{\KCmd+u}				{if {[undo [set window [ActiveWindowOrBeep]]]!=0} {HomeWindowToSelectionStart $window} else {beep}}
	addmenu {Edit} LASTCHILD 1 "Redo"							{\KCmd+y}				{if {[redo [set window [ActiveWindowOrBeep]]]!=0} {HomeWindowToSelectionStart $window} else {beep}}
	addmenu {Edit} LASTCHILD 1 "Flush Undo/Redo Buffer"			{}						{okcanceldialog "Really flush undos for:\n'[set window [ActiveWindowOrBeep]]'?";flushundos $window}
	addmenu {Edit} LASTCHILD 0 "space0" 						{\S}					{}
	addmenu {Edit} LASTCHILD 1 "Cut"							{\KCmd+x}				{SmartCut [set window [ActiveWindowOrBeep]] [CurrentClipboardOrBeep];HomeWindowToSelectionStart $window;flushundos [getclipboard]}
	addmenu {Edit} LASTCHILD 1 "Copy"							{\KCmd+c}				{SmartCopy [ActiveWindowOrBeep] [CurrentClipboardOrBeep];flushundos [getclipboard]}
	addmenu {Edit} LASTCHILD 1 "Paste"							{\KCmd+v}				{HomeWindowToSelectionStart [set window [ActiveWindowOrBeep]];paste $window [CurrentClipboardOrBeep];HomeWindowToSelectionStart $window -lenient}
	addmenu {Edit} LASTCHILD 1 "Clear"							{}						{clear [set window [ActiveWindowOrBeep]];HomeWindowToSelectionStart $window}
	addmenu {Edit} LASTCHILD 0 "space1" 						{\S}					{}
	addmenu {Edit} LASTCHILD 1 "Clipboards"						{}						{}
		addmenu {Edit Clipboards} LASTCHILD 1 "Show Current Clipboard..."	{}			{ShowCurrentClipboard}
		addmenu {Edit Clipboards} LASTCHILD 1 "Open Current Clipboard..."	{}			{OpenDefaultWindow [CurrentClipboardOrBeep]}
		addmenu {Edit Clipboards} LASTCHILD 0 "space0"			{\S}					{}
		addmenu {Edit Clipboards} LASTCHILD 1 "Clipboard 0"		{\KCmd+0}				{setclipboard clip0}
		addmenu {Edit Clipboards} LASTCHILD 1 "Clipboard 1"		{\KCmd+1}				{setclipboard clip1}
		addmenu {Edit Clipboards} LASTCHILD 1 "Clipboard 2"		{\KCmd+2}				{setclipboard clip2}
		addmenu {Edit Clipboards} LASTCHILD 1 "Clipboard 3"		{\KCmd+3}				{setclipboard clip3}
		addmenu {Edit Clipboards} LASTCHILD 1 "Clipboard 4"		{\KCmd+4}				{setclipboard clip4}
		addmenu {Edit Clipboards} LASTCHILD 1 "Clipboard 5"		{\KCmd+5}				{setclipboard clip5}
		addmenu {Edit Clipboards} LASTCHILD 1 "Clipboard 6"		{\KCmd+6}				{setclipboard clip6}
		addmenu {Edit Clipboards} LASTCHILD 1 "Clipboard 7"		{\KCmd+7}				{setclipboard clip7}
		addmenu {Edit Clipboards} LASTCHILD 1 "Clipboard 8"		{\KCmd+8}				{setclipboard clip8}
		addmenu {Edit Clipboards} LASTCHILD 1 "Clipboard 9"		{\KCmd+9}				{setclipboard clip9}
	addmenu {Edit} LASTCHILD 0 "space2" 						{\S}					{}
	addmenu {Edit} LASTCHILD 1 "Strip EOL Whitespace"			{}						{set numRemoved [StripWhite [set window [ActiveWindowOrBeep]]];okdialog "$window\n\nCharacters removed: $numRemoved"}
	addmenu {Edit} LASTCHILD 1 "Set Unix LF Line Termination"	{}						{SetLineTermination [ActiveWindowOrBeep] "\n"}
	addmenu {Edit} LASTCHILD 1 "Set PC CRLF Line Termination"	{}						{SetLineTermination [ActiveWindowOrBeep] "\r\n"}

	addmenu {Edit} LASTCHILD 1 "In All Open Windows ..."		{}						{}
	addmenu {Edit "In All Open Windows ..."} LASTCHILD 1 "Strip EOL Whitespace"			{}	{ActiveWindowOrBeep;set numRemoved 0;foreach window [windowlist] {incr numRemoved [StripWhite $window]};okdialog "Total characters removed: $numRemoved"}
	addmenu {Edit "In All Open Windows ..."} LASTCHILD 1 "Set Unix Line Termination"	{}	{ActiveWindowOrBeep;foreach window [windowlist] {SetLineTermination $window "\n"};okdialog "All windows completed"}
	addmenu {Edit "In All Open Windows ..."} LASTCHILD 1 "Set PC Line Termination"		{}	{ActiveWindowOrBeep;foreach window [windowlist] {SetLineTermination $window "\r\n"};okdialog "All windows completed"}

	addmenu {Edit} LASTCHILD 0 "space3" 						{\S}					{}
	addmenu {Edit} LASTCHILD 1 "Insert Unicode"					{\KCmd+Shift+u}			{InsertUnicode [ActiveWindowOrBeep]}
	addmenu {Edit} LASTCHILD 0 "space4" 						{\S}					{}
	addmenu {Edit} LASTCHILD 1 "Align Left"						{\KCmd+Shift+bracketleft}	{AlignLeft [ActiveWindowOrBeep]}
	addmenu {Edit} LASTCHILD 1 "Shift Left"						{\KCmd+bracketleft}	{ShiftLeft [ActiveWindowOrBeep]}
	addmenu {Edit} LASTCHILD 1 "Shift Right"					{\KCmd+bracketright}	{ShiftRight [ActiveWindowOrBeep]}
	addmenu {Edit} LASTCHILD 1 "Upper case Selection"			{}						{UpperCaseSelection [ActiveWindowOrBeep]}
	addmenu {Edit} LASTCHILD 1 "Lower case Selection"			{}						{LowerCaseSelection [ActiveWindowOrBeep]}
	addmenu {Edit} LASTCHILD 1 "Increment Selected Numbers..."	{}						{IncrementSelection [ActiveWindowOrBeep] [textdialog "How much to increment:" 1]}
	addmenu {Edit} LASTCHILD 1 "Enumerate Selections..."		{}						{EnumerateSelection [ActiveWindowOrBeep] [textdialog "Starting value:" 1] [textdialog "How much to increment:" 1]}
	addmenu {Edit} LASTCHILD 1 "Sum of Selected Numbers"		{}						{SumSelection [ActiveWindowOrBeep]}
	addmenu {Edit} LASTCHILD 1 "Sort Selection Segments"		{}						{SortSelection [ActiveWindowOrBeep]}
	addmenu {Edit} LASTCHILD 1 "Reverse Selection Segments"		{}						{ReverseSelection [ActiveWindowOrBeep]}
	addmenu {Edit} LASTCHILD 1 "Refill Selection..."			{\KCmd+r}				{RefillSelection [ActiveWindowOrBeep]}

addmenu {} LASTCHILD 1 "Selection" "" ""
	addmenu {Selection} LASTCHILD 1 "Select All"				{\KCmd+a}				{selectall [ActiveWindowOrBeep]}
	addmenu {Selection} LASTCHILD 0 "space0"					{\S}					{}
	addmenu {Selection} LASTCHILD 1 "Invert"					{}						{openmark [set window [ActiveWindowOrBeep]] _seltemp;selectall $window -mark _seltemp;xorselection $window -sourcemark _seltemp;closemark $window _seltemp}
	addmenu {Selection} LASTCHILD 0 "space1"					{\S}					{}
	addmenu {Selection} LASTCHILD 1 "Select Between Braces"		{\KCmd+i}				{SelectLinesBetweenBraces [ActiveWindowOrBeep]}
	addmenu {Selection} LASTCHILD 0 "space2"					{\S}					{}
	addmenu {Selection} LASTCHILD 1 "Unselect Whitespace"		{\KCmd+b}				{UnselectWhitespace [ActiveWindowOrBeep]}
	addmenu {Selection} LASTCHILD 1 "Unselect Non-Identifier"	{\KCmd+Shift+b}			{UnselectNonIdentifier [ActiveWindowOrBeep]}
	addmenu {Selection} LASTCHILD 0 "space3" 					{\S}					{}
	addmenu {Selection} LASTCHILD 1 "Mark..."					{}						{SetMark [ActiveWindowOrBeep] [textdialog "Mark selection with what name?"]}
	addmenu {Selection} LASTCHILD 1 "Unmark..."					{}						{foreach mark [listdialog "Delete which mark(s):" [marklist [set window [ActiveWindowOrBeep]]]] {closemark $window $mark}}
	addmenu {Selection} LASTCHILD 1 "Goto Mark..."				{}						{foreach mark [listdialog "Go to which mark:" [marklist [set window [ActiveWindowOrBeep]]]] {GotoMark $window $mark; HomeWindowToSelection $window -strict}}
	addmenu {Selection} LASTCHILD 0 "space4" 					{\S}					{}
	addmenu {Selection} LASTCHILD 1 "Green Mark..."				{}						{}
	addmenu {Selection "Green Mark..."} LASTCHILD 1 "Set"		{\KCmd+F1}				{SetMark [set window [ActiveWindowOrBeep]] F1;markvisible $window F1 008000}
	addmenu {Selection "Green Mark..."} LASTCHILD 1 "Add"		{\KCmd+Shift+F1}		{orselection [ActiveWindowOrBeep] -destmark F1}
	addmenu {Selection "Green Mark..."} LASTCHILD 1 "Goto"		{\KF1}					{GotoMark [set window [ActiveWindowOrBeep]] F1;HomeWindowToSelection $window -semistrict}
	addmenu {Selection} LASTCHILD 1 "Red Mark..."				{}						{}
	addmenu {Selection "Red Mark..."} LASTCHILD 1 "Set"			{\KCmd+F2}				{SetMark [set window [ActiveWindowOrBeep]] F2;markvisible $window F2 800000}
	addmenu {Selection "Red Mark..."} LASTCHILD 1 "Add"			{\KCmd+Shift+F2}		{orselection [ActiveWindowOrBeep] -destmark F2}
	addmenu {Selection "Red Mark..."} LASTCHILD 1 "Goto"		{\KF2}					{GotoMark [set window [ActiveWindowOrBeep]] F2;HomeWindowToSelection $window -semistrict}
	addmenu {Selection} LASTCHILD 1 "Blue Mark..."				{}						{}
	addmenu {Selection "Blue Mark..."} LASTCHILD 1 "Set"		{\KCmd+F3}				{SetMark [set window [ActiveWindowOrBeep]] F3;markvisible $window F3 6060A0}
	addmenu {Selection "Blue Mark..."} LASTCHILD 1 "Add"		{\KCmd+Shift+F3}		{orselection [ActiveWindowOrBeep] -destmark F3}
	addmenu {Selection "Blue Mark..."} LASTCHILD 1 "Goto"		{\KF3}					{GotoMark [set window [ActiveWindowOrBeep]] F3;HomeWindowToSelection $window -semistrict}
	addmenu {Selection} LASTCHILD 1 "Teal Mark..."				{}						{}
	addmenu {Selection "Teal Mark..."} LASTCHILD 1 "Set"		{\KCmd+F4}				{SetMark [set window [ActiveWindowOrBeep]] F4;markvisible $window F4 007070}
	addmenu {Selection "Teal Mark..."} LASTCHILD 1 "Add"		{\KCmd+Shift+F4}		{orselection [ActiveWindowOrBeep] -destmark F4}
	addmenu {Selection "Teal Mark..."} LASTCHILD 1 "Goto"		{\KF4}					{GotoMark [set window [ActiveWindowOrBeep]] F4;HomeWindowToSelection $window -semistrict}

addmenu {} LASTCHILD 1 "Find" "" ""
	addmenu {Find} LASTCHILD 1 "Find/Replace..."				{\KCmd+f}				{AskSearch [ActiveWindowOrBeep]}
	addmenu {Find} LASTCHILD 1 "Find/Replace in All..."			{}						{ActiveWindowOrBeep;AskSearchAll}
	addmenu {Find} LASTCHILD 1 "Find Same"						{\KCmd+g}				{FindNext [ActiveWindowOrBeep] 0}
	addmenu {Find} LASTCHILD 1 "Find Same Backwards"			{\KCmd+Shift+g}			{FindNext [ActiveWindowOrBeep] 1}
	addmenu {Find} LASTCHILD 1 "Find Selection"					{\KCmd+h}				{FindSelectionNext [ActiveWindowOrBeep] 0}
	addmenu {Find} LASTCHILD 1 "Find Selection Backwards"		{\KCmd+Shift+h}			{FindSelectionNext [ActiveWindowOrBeep] 1}
	addmenu {Find} LASTCHILD 0 "space0" 						{\S}					{}
	addmenu {Find} LASTCHILD 1 "Replace Same"					{\KCmd+t}				{ReplaceNext [ActiveWindowOrBeep] 0}
	addmenu {Find} LASTCHILD 1 "Replace Same Backwards"			{\KCmd+Shift+t}			{ReplaceNext [ActiveWindowOrBeep] 1}
	addmenu {Find} LASTCHILD 0 "space1" 						{\S}					{}
	addmenu {Find} LASTCHILD 1 "Replace Selections With..."		{\KCmd+Shift+r}			{ReplaceSelections [ActiveWindowOrBeep]}
	addmenu {Find} LASTCHILD 0 "space2" 						{\S}					{}
	addmenu {Find} LASTCHILD 1 "Go To Line..."					{\KCmd+l}				{selectline [set window [ActiveWindowOrBeep]] [expr [textdialog "Go to line:"]];HomeWindowToSelectionStart $window}
	addmenu {Find} LASTCHILD 1 "Locate Selection"				{\KCmd+Shift+l}			{HomeWindowToSelection [ActiveWindowOrBeep] -strict}
	addmenu {Find} LASTCHILD 0 "space3" 						{\S}					{}
	addmenu {Find} LASTCHILD 1 "Grep -n For..."					{\KCmd+Shift+f}			{set data "grep -E -I -D skip -n -s --line-buffered [set grepFilesChoice [textdialog "Enter pattern, and file list: (eg. void *.c)" $grepFilesChoice]]"; set window [NewTitledWindow "Search '$grepFilesChoice'"]; task $window $data}

addmenu {} LASTCHILD 1 "Window" "" ""
	addmenu {Window} LASTCHILD 1 "Encoding"						{}						{}
	addmenu {Window Encoding} LASTCHILD 1 "ASCII"				{}						{setutf8 [ActiveWindowOrBeep] 0}
	addmenu {Window Encoding} LASTCHILD 1 "UTF-8"				{}						{setutf8 [ActiveWindowOrBeep] 1}
	addmenu {Window Encoding} LASTCHILD 1 "Toggle Encoding"		{\KF12}					{setutf8 [ActiveWindowOrBeep] [expr [getutf8 [ActiveWindowOrBeep]]^1]}
	addmenu {Window} LASTCHILD 0 "space0"						{\S}					{}
	addmenu {Window} LASTCHILD 1 "Set Tab Size..."				{\KCmd+e}				{settabsize [set window [ActiveWindowOrBeep]] [expr [textdialog "New tab size:" [gettabsize $window]]]}
	addmenu {Window} LASTCHILD 0 "space1"						{\S}					{}
	addmenu {Window} LASTCHILD 1 "Set Font..."					{\KCmd+Shift+e}			{setfont [set window [ActiveWindowOrBeep]] [fontdialog "Choose Font:" [getfont $window]]}
	addmenu {Window} LASTCHILD 1 "Decrease Font Size"			{\KCmd+minus}			{SetFontSize [set window [ActiveWindowOrBeep]] [expr [GetFontSize $window]-1]}
	addmenu {Window} LASTCHILD 1 "Increase Font Size"			{\KCmd+equal}			{SetFontSize [set window [ActiveWindowOrBeep]] [expr [GetFontSize $window]+1]}
	addmenu {Window} LASTCHILD 0 "space2"						{\S}					{}
	addmenu {Window} LASTCHILD 1 "Swap Top Windows"				{\KCmd+space}			{SwapWindows}
	addmenu {Window} LASTCHILD 1 "Rotate Windows"				{\KCmd+grave}			{RotateWindows}
	addmenu {Window} LASTCHILD 1 "Choose Window..."				{\KCmd+Shift+o}			{ChooseWindow}
	addmenu {Window} LASTCHILD 0 "space3"						{\S}					{}
	addmenu {Window} LASTCHILD 1 "Stack Windows"				{}						{StackWindows}
	addmenu {Window} LASTCHILD 1 "Tile Windows"					{}						{TileWindows}
	addmenu {Window} LASTCHILD 0 "space4"						{\S}					{}
	addmenu {Window} LASTCHILD 1 "Language Mode"				{}						{}
		addmenu {Window "Language Mode"} LASTCHILD 1 "<none>"	{}						{setsyntaxmap [ActiveWindowOrBeep] ""}
		foreach element [lsort -dictionary [syntaxmaps]] \
		{
			addmenu {Window "Language Mode"} LASTCHILD 1 $element	{}					"setsyntaxmap \[ActiveWindowOrBeep\] \"$element\""
		}
	addmenu {Window} LASTCHILD 1 "Color Scheme"						{}					{}
		foreach element [lsort -dictionary [FilterSystemElements [colorschemes]]] \
		{
			addmenu {Window "Color Scheme"} LASTCHILD 1 $element	{}					"setcolorscheme \[ActiveWindowOrBeep\] \"$element\""
		}

addmenu {} LASTCHILD 1 "Misc" "" ""
	addmenu {Misc} LASTCHILD 1 "Line Wrapping..."				{}						{WrapOnOff}
	addmenu {Misc} LASTCHILD 1 "Get man Page..."				{\KCmd+slash}			{ManPage [textdialog "Enter man page subject:"]}
	addmenu {Misc} LASTCHILD 1 "Spell"							{}						{SpellDocument [ActiveWindowOrBeep]}
	addmenu {Misc} LASTCHILD 1 "Mail To..."						{}						{MailDocument [ActiveWindowOrBeep]}
	addmenu {Misc} LASTCHILD 1 "Mail Windows To..."				{}						{MailWindows}

addmenu {} LASTCHILD 1 "Tasks" "" ""
	addmenu {Tasks} LASTCHILD 1 "Execute Shell Task"			{\KCmd+Return}			{task [set window [ActiveWindowOrBeep]] [EvalText $window]}
	addmenu {Tasks} LASTCHILD 1 "Send EOF To Task"				{\KCmd+k}				{eoftask [ActiveWindowOrBeep]}
	addmenu {Tasks} LASTCHILD 1 "Kill Task"						{\KCmd+Shift+k}			{killtask [ActiveWindowOrBeep]}
	addmenu {Tasks} LASTCHILD 0 "space0"						{\S}					{}
	addmenu {Tasks} LASTCHILD 1 "Pipe Selections Through..."	{}						{PipeSelection [ActiveWindowOrBeep] [PipeCommandGet "Enter command to pipe selections through:"] }
	addmenu {Tasks} LASTCHILD 1 "Pipe Windows Through..."		{}						{ActiveWindowOrBeep;set command [PipeWindowCommandGet "Enter command to pipe windows through:"];foreach window [windowlist] {PipeWindow $window $command};okdialog "All windows completed"}

addmenu {} LASTCHILD 1 "Build" "" ""
	addmenu {Build} LASTCHILD 1 "make"							{\KCmd+m}				{HomeWindowToSelectionStart [set window [ActiveWindowOrBeep]];breakundo $window;task $window make}

# Bind keys to various useful things.

bindkey {F5}					{WrapOnOff};
bindkey {Cmd+F5}				{WrapNewColumn};

# bind Cmd+keypad enter to run a shell command just like the menu item for the enter key
bindkey {Cmd+KP_Enter}			{task [set window [ActiveWindowOrBeep]] [EvalText $window]}
# bind Cmd-shift enter, and kp_enter to run Tcl
bindkey {Cmd+Shift+Return}		{execute [set window [ActiveWindowOrBeep]] [EvalText $window];catch {HomeWindowToSelectionStart $window -lenient}}
bindkey {Cmd+Shift+KP_Enter}	{execute [set window [ActiveWindowOrBeep]] [EvalText $window];catch {HomeWindowToSelectionStart $window -lenient}}

bindkey {Help}					{ManPage [textdialog "Enter man page subject:"]}

# Specify initial paths to be placed into the directory menu.

AddDirectoryMenu /
catch {AddDirectoryMenu [glob ~]};		# this can fail if user has no home directory
AddDirectoryMenu [pwd]


# see if the user has his own set of modules specified, if so, skip including the defaults, and get only his
if {[file exists [file join $::userFilesDir $::modulesDir]]} \
{
	foreach file [lsort [glob -nocomplain [file join $::userFilesDir $::modulesDir *.tcl]]] \
	{
		source $file;
	}
} \
else \
{
	# use all system defined modules
	foreach file [lsort [glob -nocomplain [file join $::sysFilesDir $::modulesDir *.tcl]]] \
	{
		source $file;
	}
}

# get user's additional modules (if any)
if {[file exists [file join $::userFilesDir $::auxModulesDir]]} \
{
	foreach file [lsort [glob -nocomplain [file join $::userFilesDir $::auxModulesDir *.tcl]]] \
	{
		source $file;
	}
}
