// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// view tracking modes
#define	TM_mTRACK_MODE		0x000F		// tells what the pointer track is doing
#define	TM_NORMAL			0x0000		// tracking normal selection
#define	TM_COLUMNAR			0x0001		// tracking columnar selection
#define	TM_HAND				0x0002		// tracking "hand" movement
#define	TM_mSELECT_MODE		0x00F0		// masks out selection modes
#define	TM_CHAR				0x0000		// select by character
#define	TM_WORD				0x0010		// select by word
#define	TM_LINE				0x0020		// select by line
#define	TM_ALL				0x0030		// select everything
#define	TM_mSELECT_MATH		0x0F00		// tells what math to apply to selection
#define	TM_REPLACE			0x0000		// selection is being replaced
#define	TM_ADD				0x0100		// selection is being added to
#define	TM_SUBTRACT			0x0200		// selection is being subtracted from
#define	TM_XOR				0x0300		// selection is being XOR'd
#define	TM_mCONTINUE		0x1000		// tells if selection should be continued from previous
#define	TM_mREPEAT			0x8000		// flag that tells if this is a repeat (mouse still held)

// relative position modes (used to move cursor, delete, etc...)

#define	RPM_mMODE			0x000F		// masks out position modes
#define	RPM_CHAR			0x0000
#define	RPM_WORD			0x0001
#define	RPM_LINE			0x0002
#define	RPM_LINEEDGE		0x0003
#define	RPM_PARAGRAPHEDGE	0x0004
#define	RPM_PAGE			0x0005
#define	RPM_DOCEDGE			0x0006
#define	RPM_mBACKWARD		0x8000		// if set, move backward (towards start of file), if clear, move forward

uint64_t EditorOffsetFix(uint64_t offsetToFix,uint64_t startOffset,uint64_t oldEndOffset,uint64_t newEndOffset);
void EditorViewToRangeHome(EDITOR_VIEW *view,uint64_t startPosition,uint64_t endPosition,bool useEnd,int horizontalMovement,int verticalMovement);
void EditorViewToSelectionHome(EDITOR_VIEW *view,bool useEnd,int horizontalMovement,int verticalMovement);
void EditorViewToSelectionEdgeHome(EDITOR_VIEW *view,bool useEnd,int horizontalMovement,int verticalMovement);
void EditorVerticalScroll(EDITOR_VIEW *view,int64_t amountToScroll);
void EditorVerticalScrollByPages(EDITOR_VIEW *view,int64_t pagesToScroll);
void EditorHorizontalScroll(EDITOR_VIEW *view,int64_t amountToScroll);
void EditorHorizontalScrollByPages(EDITOR_VIEW *view,int64_t pagesToScroll);
void EditorSelectionChangeEnd(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse);
void EditorSelectionChangeStart(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse);
bool EditorTextReplaceWithStream(EDITOR_BUFFER *buffer,uint64_t startOffset,uint64_t length,TEXT_STREAM *stream,uint64_t numBytes,bool allowUndoMerge);
bool EditorTextReplace(EDITOR_BUFFER *buffer,uint64_t startOffset,uint64_t length,uint8_t *text,uint64_t numBytes,bool allowUndoMerge);
bool EditorFileReplace(EDITOR_BUFFER *buffer,uint64_t startOffset,uint64_t length,const char *path);
void EditorReplaceEnd(EDITOR_BUFFER *buffer);
void EditorReplaceStart(EDITOR_BUFFER *buffer);
bool EditorSelectedTextDelete(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse);
void EditorSelectionInfoGet(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,uint64_t *startPosition,uint64_t *endPosition,uint64_t *startLine,uint64_t *endLine,uint64_t *startLinePosition,uint64_t *endLinePosition,uint64_t *totalSegments,uint64_t *totalSpan);
uint8_t *EditorNextSelectionToBuffer(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,uint64_t *currentPosition,uint64_t additionalLength,uint64_t *actualLength,bool *atEnd);
bool EditorSelectionDelete(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse);
bool EditorSelectionSet(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,uint64_t startPosition,uint64_t endPosition);
void EditorViewPointerTrackComplete(EDITOR_VIEW *view);
bool EditorViewPointerTrack(EDITOR_VIEW *view,int64_t viewX,int64_t viewLine,uint16_t trackMode);
bool EditorCursorMove(EDITOR_VIEW *view,SELECTION_UNIVERSE *selectionUniverse,uint16_t relativeMode);
bool EditorSelectionMove(EDITOR_VIEW *view,SELECTION_UNIVERSE *selectionUniverse,uint16_t relativeMode);
bool EditorNormalSelectionExpand(EDITOR_VIEW *view,SELECTION_UNIVERSE *selectionUniverse,uint16_t relativeMode);
bool EditorNormalSelectionReduce(EDITOR_VIEW *view,SELECTION_UNIVERSE *selectionUniverse,uint16_t relativeMode);
bool EditorLineLocate(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,uint64_t line);
bool EditorSelectAll(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse);
bool EditorDelete(EDITOR_VIEW *view,uint16_t relativeMode);
bool EditorInsert(EDITOR_BUFFER *buffer,uint8_t *text,uint64_t textLength);
bool EditorAuxInsert(EDITOR_BUFFER *buffer,uint8_t *text,uint64_t textLength);
bool EditorFileInsert(EDITOR_BUFFER *buffer,const char *path);
bool EditorAutoIndent(EDITOR_BUFFER *buffer);
void EditorBufferViewUnlink(EDITOR_VIEW *view);
void EditorBufferViewLink(EDITOR_BUFFER *buffer,EDITOR_VIEW *view);
