// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

typedef void DIALOG_POSITION_FUNCTION(struct dialogItem *item,EDITOR_RECT *rect);	// called by the item to get its position within the dialog

typedef struct dialogItem	// item in list of items for given dialog (built during dialog creation)
{
	struct dialog
		*dialog;			// pointer to the dialog this item is in
	void
		(*disposeFunction)(struct dialogItem *item);				// routine to call to get rid of item's local storage
	void
		(*drawFunction)(struct dialogItem *item,GUI_IMAGE *imageBuffer,Region invalidRegion);	// routine to call when drawing item
	bool
		(*trackFunction)(struct dialogItem *item,XEvent *event);	// routine to call when attempting to track the item (returns false if initial point not in item)
	void
		(*resizeFunction)(struct dialogItem *item);					// call this after the window has changed size to resize/reposition the item
	bool
		(*earlyKeyFunction)(struct dialogItem *item,EDITOR_KEY *editorKey);	// if this item wants certain keys before the default item can get them, it defines this
	bool
		wantFocus;													// tells the dialog manager that this item desires to be in the focus chain
	void
		(*focusChangeFunction)(struct dialogItem *item);			// called by dialog manager to tell item that focus is changing
	bool
		(*focusKeyFunction)(struct dialogItem *item,EDITOR_KEY *editorKey);	// when keys come in for this focussed item, this is called -- it returns true if the key was handled
	void
		(*focusPeriodicFunction)(struct dialogItem *item);			// if this item wants to flash the cursor or whatever, it can set up this routine
	struct dialogItem
		*previousItem,
		*nextItem;
} DIALOG_ITEM;

typedef struct dialog	DIALOG;

void DialogFocusTake(DIALOG_ITEM *item);
DIALOG_ITEM *DialogFocusItemGet(DIALOG *dialog);
EDITOR_WINDOW *DialogWindowGet(DIALOG *dialog);
void DialogColorsGet(EDITOR_COLOR *backgroundColor);
void DialogColorsSet(EDITOR_COLOR backgroundColor);
void DialogStaticTextColorsGet(EDITOR_COLOR *foregroundColor);
void DialogStaticTextColorsSet(EDITOR_COLOR foregroundColor);
void DialogTextBoxColorsGet(EDITOR_COLOR *activeBorderColor,EDITOR_COLOR *inactiveBorderColor);
void DialogTextBoxColorsSet(EDITOR_COLOR activeBorderColor,EDITOR_COLOR inactiveBorderColor);
void DialogListBoxColorsGet(EDITOR_COLOR *activeBackgroundColor,EDITOR_COLOR *activeForegroundColor,EDITOR_COLOR *activeBorderColor,EDITOR_COLOR *inactiveBackgroundColor,EDITOR_COLOR *inactiveForegroundColor,EDITOR_COLOR *inactiveBorderColor);
void DialogListBoxColorsSet(EDITOR_COLOR activeBackgroundColor,EDITOR_COLOR activeForegroundColor,EDITOR_COLOR activeBorderColor,EDITOR_COLOR inactiveBackgroundColor,EDITOR_COLOR inactiveForegroundColor,EDITOR_COLOR inactiveBorderColor);
void DialogCheckBoxColorsGet(EDITOR_COLOR *activeBackgroundColor,EDITOR_COLOR *activeForegroundColor,EDITOR_COLOR *inactiveBackgroundColor,EDITOR_COLOR *inactiveForegroundColor);
void DialogCheckBoxColorsSet(EDITOR_COLOR activeBackgroundColor,EDITOR_COLOR activeForegroundColor,EDITOR_COLOR inactiveBackgroundColor,EDITOR_COLOR inactiveForegroundColor);
void DialogPopupColorsGet(EDITOR_COLOR *activeBackgroundColor,EDITOR_COLOR *activeForegroundColor,EDITOR_COLOR *inactiveBackgroundColor,EDITOR_COLOR *inactiveForegroundColor,EDITOR_COLOR *borderColor);
void DialogPopupColorsSet(EDITOR_COLOR activeBackgroundColor,EDITOR_COLOR activeForegroundColor,EDITOR_COLOR inactiveBackgroundColor,EDITOR_COLOR inactiveForegroundColor,EDITOR_COLOR borderColor);
void DialogButtonColorsGet(EDITOR_COLOR *activeBackgroundColor,EDITOR_COLOR *activeForegroundColor,EDITOR_COLOR *inactiveBackgroundColor,EDITOR_COLOR *inactiveForegroundColor,EDITOR_COLOR *pressedBackgroundColor,EDITOR_COLOR *pressedForegroundColor,EDITOR_COLOR *shadowColor);
void DialogButtonColorsSet(EDITOR_COLOR activeBackgroundColor,EDITOR_COLOR activeForegroundColor,EDITOR_COLOR inactiveBackgroundColor,EDITOR_COLOR inactiveForegroundColor,EDITOR_COLOR pressedBackgroundColor,EDITOR_COLOR pressedForegroundColor,EDITOR_COLOR shadowColor);
const char *DialogStaticTextFontGet();
void DialogStaticTextFontSet(const char *fontName);
const char *DialogTextBoxFontGet();
void DialogTextBoxFontSet(const char *fontName);
const char *DialogListBoxFontGet();
void DialogListBoxFontSet(const char *fontName);
const char *DialogCheckBoxFontGet();
void DialogCheckBoxFontSet(const char *fontName);
const char *DialogPopupFontGet();
void DialogPopupFontSet(const char *fontName);
const char *DialogButtonFontGet();
void DialogButtonFontSet(const char *fontName);
bool PlatformDialogOk(const char *text);
bool PlatformDialogOkCancel(const char *text,bool *cancel);
bool PlatformDialogYesNoCancel(const char *text,bool *yes,bool *cancel);
bool PlatformDialogSimpleText(const char *title,char *enteredText,uint32_t stringBytes,bool *cancel);
bool PlatformDialogSearchReplace(EDITOR_BUFFER *searchBuffer,EDITOR_BUFFER *replaceBuffer,bool *backwards,bool *wrapAround,bool *ignoreCase,bool *limitScope,bool *selectionExpr,bool *useSubexpression,uint32_t *subexpressionIndex,bool *replaceFunction,uint16_t *searchType,bool *cancel);
bool PlatformDialogSimpleList(const char *title,uint32_t numElements,const char **listElements,bool *selectedElements,bool *cancel);
bool PlatformDialogOpenFile(const char *title,char *fullPath,uint32_t stringBytes,char ***listElements,bool *cancel);
void PlatformDialogFreeOpenFilePaths(char **paths);
bool PlatformDialogSaveFile(const char *title,char *fullPath,uint32_t stringBytes,bool *cancel);
bool PlatformDialogChoosePath(const char *title,char *fullPath,uint32_t stringBytes,bool *cancel);
bool PlatformDialogChooseFont(const char *title,char *font,uint32_t stringBytes,bool *cancel);
void PlatformDialogCenterPointSet(int32_t centerX,int32_t centerY);
void DialogUnInit();
bool DialogInit();
