// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// Generic strucutre that manages image data
// This helps hide the numerous X11 and extension concepts
// from the rest of the code.
typedef struct guiImage
{
	uint32_t
		width,
		height;
	Pixmap
		pixmap;		// the server pixmap holding the data
	XftDraw
		*xftDraw;	// an Xft handle to the pixmap
	Picture
		picture;	// and Xrender handle to the pixmap
} GUI_IMAGE;

void ImageEditorColorToRenderColorAlpha(EDITOR_COLOR color,XRenderColor *renderColor);
void ImageEditorColorToRenderColor(EDITOR_COLOR color,XRenderColor *renderColor);
void ImageClipRegionSet(GUI_IMAGE *image,Region region);
void ImageClipRegionRectSet(GUI_IMAGE *image,EDITOR_RECT *rect);
void ImageDestroy(GUI_IMAGE *image);
GUI_IMAGE *ImageCreate(uint32_t width,uint32_t height);
GUI_IMAGE *ImageRead(const char *path);
void ImageDraw(const GUI_IMAGE *source,GUI_IMAGE *destination,int32_t destX,int32_t destY);
void ImageDrawMask(const GUI_IMAGE *source,EDITOR_COLOR color,GUI_IMAGE *destination,int32_t destX,int32_t destY);
void ImageDrawToWindow(const GUI_IMAGE *source,Drawable destination,Region invalidRegion,uint32_t width,uint32_t height);
void ImageFillAlpha(GUI_IMAGE *dest,EDITOR_COLOR color,int32_t x,int32_t y,uint32_t w,uint32_t h);
void ImageFill(GUI_IMAGE *dest,EDITOR_COLOR color,int32_t x,int32_t y,uint32_t w,uint32_t h);
void ImageUnInit();
bool ImageInit();
