// Manage image assets used by the editor.
// Copyright © 2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define		IMAGE_DIRECTORY					LIB_PATH_PREFIX "/images"		// directory where image data is loaded from

#define		ICON_IMAGE_FILE					"icon.png"

GUI_IMAGE
	*leftArrowImage,
	*rightArrowImage,
	*upArrowImage,
	*downArrowImage,
	*buttonImage,
	*checkBoxImage,
	*checkedBoxImage,
	*subMenuImage,
	*bufferModifiedImage,
	*taskRunningImage;

typedef struct
{
	GUI_IMAGE
		**image;
	const char
		*name;
} IMAGE_LOAD;

static const IMAGE_LOAD
	imageLoad[]=
	{
		{&leftArrowImage,		"leftArrow.png"},
		{&rightArrowImage,		"rightArrow.png"},
		{&upArrowImage,			"upArrow.png"},
		{&downArrowImage,		"downArrow.png"},
		{&buttonImage,			"button.png"},
		{&checkBoxImage,		"checkBox.png"},
		{&checkedBoxImage,		"checkedBox.png"},
		{&subMenuImage,			"subMenu.png"},
		{&bufferModifiedImage,	"bufferModified.png"},
		{&taskRunningImage,		"taskRunning.png"},
	};

static unsigned long
	*e93Icon;				// pointer to icon read at startup (unsigned long type is required here)
static uint32_t
	e93IconElements;		// number of unsigned longs used in the icon (needed for _NET_WM_ICON)

unsigned char *ImageAssetIconGet(uint32_t *numElements)
// Get the icon image data that was read at startup.
{
	*numElements=e93IconElements;
	return((unsigned char *)e93Icon);
}

void ImageAssetUnInit()
// Undo what ImageAssetInit did.
{
	uint32_t
		imageIndex;

	imageIndex=elementsof(imageLoad);
	while(imageIndex)
	{
		imageIndex--;
		ImageDestroy(*(imageLoad[imageIndex].image));
	}
	ImageFileCardinal32Destroy(e93Icon);
}

bool ImageAssetInit()
// Read all the image assets needed by the editor at startup.
{
	bool
		fail;
	uint32_t
		imageIndex;
	char
		completePath[1024];

	if((e93Icon=ImageFileCardinal32Read(IMAGE_DIRECTORY "/" ICON_IMAGE_FILE,&e93IconElements)))	// get the icon
	{
		imageIndex=0;
		fail=false;
		while((!fail)&&(imageIndex<elementsof(imageLoad)))
		{
			snprintf(completePath,sizeof(completePath),"%s/%s",IMAGE_DIRECTORY,imageLoad[imageIndex].name);
			fail=!((*(imageLoad[imageIndex].image))=ImageRead(completePath));
			imageIndex++;
		}
		if(!fail)
		{
			return(true);
		}
		imageIndex--;			// step back for the load that failed
		while(imageIndex)
		{
			imageIndex--;
			ImageDestroy(*(imageLoad[imageIndex].image));
		}
		ImageFileCardinal32Destroy(e93Icon);
	}
	return(false);
}
