// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// defines and data structures for scrollBars

#define	SCROLL_BAR_WIDTH	20									// number of pixels wide that the scroll bars are
#define	ARROW_LENGTH		20									// number of pixels long that the arrows are (if possible)
#define	THUMB_LENGTH		20									// minimum number of pixels long that the thumb is (if possible)

typedef struct scrollBar
{
	EDITOR_WINDOW
		*parentWindow;											// the window that contains this scroll bar
	bool
		horizontal;
	int32_t
		x,
		y;
	uint32_t
		length;													// length in pixels of the scroll bar
	uint64_t
		thumbMax;												// indicates the maximum number of positions for the thumb
	uint64_t
		thumbPosition;											// indicates the current position of the thumb
	bool
		highlightLowArrow,										// used during tracking to remember what is highlighted
		highlightHighArrow,
		highlightLowPage,
		highlightHighPage,
		highlightThumb;
	void
		(*thumbProc)(struct scrollBar *scrollBar,uint64_t newPosition,void *parameters);	// routine called when scroll bar is tracking the thumb
	void
		(*stepLowerProc)(struct scrollBar *scrollBar,bool repeating,void *parameters);	// routine called when scroll bar is stepping to lower values
	void
		(*stepHigherProc)(struct scrollBar *scrollBar,bool repeating,void *parameters);	// routine called when scroll bar is stepping to higher values
	void
		(*pageLowerProc)(struct scrollBar *scrollBar,bool repeating,void *parameters);	// routine called when scroll bar is paging to lower values
	void
		(*pageHigherProc)(struct scrollBar *scrollBar,bool repeating,void *parameters);	// routine called when scroll bar is paging to higher values
	void
		*procParameters;										// data passed to routines
} SCROLL_BAR;

typedef struct scrollBarDescriptor								// describes a scroll bar to creation routines
{
	bool
		horizontal;
	int32_t
		x,
		y;
	uint32_t
		length;													// length in pixels of the scroll bar
	uint64_t
		thumbMax;												// indicates the maximum number of positions for the thumb
	uint64_t
		thumbPosition;											// indicates the current position of the thumb
	void
		(*thumbProc)(struct scrollBar *scrollBar,uint64_t newPosition,void *parameters);	// routine called when scroll bar is tracking the thumb
	void
		(*stepLowerProc)(struct scrollBar *scrollBar,bool repeating,void *parameters);	// routine called when scroll bar is stepping to lower values
	void
		(*stepHigherProc)(struct scrollBar *scrollBar,bool repeating,void *parameters);	// routine called when scroll bar is stepping to higher values
	void
		(*pageLowerProc)(struct scrollBar *scrollBar,bool repeating,void *parameters);	// routine called when scroll bar is paging to lower values
	void
		(*pageHigherProc)(struct scrollBar *scrollBar,bool repeating,void *parameters);	// routine called when scroll bar is paging to higher values
	void
		*procParameters;										// data passed to routines
} SCROLL_BAR_DESCRIPTOR;


bool ScrollBarPointInside(SCROLL_BAR *scrollBar,int32_t x,int32_t y);
void ScrollBarDraw(SCROLL_BAR *scrollBar,GUI_IMAGE *imageBuffer,Region invalidRegion);
bool ScrollBarTrack(SCROLL_BAR *scrollBar,XEvent *event);
void ScrollBarReposition(SCROLL_BAR *scrollBar,int32_t x,int32_t y,uint32_t length,bool horizontal);
void ScrollBarAdjust(SCROLL_BAR *scrollBar,uint64_t thumbMax,uint64_t thumbPosition);
void ScrollBarDispose(SCROLL_BAR *scrollBar);
SCROLL_BAR *ScrollBarCreate(EDITOR_WINDOW *window,SCROLL_BAR_DESCRIPTOR *description);
