// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

typedef struct platformViewInfo
{
	Window
		cursorWindow;					// x InputOnly window used to make the cursor change to caret within the view (this is ALL it is used for!)
	EDITOR_RECT
		bounds;							// boundary of this view within its parent window (all drawing happens within these bounds)
	int64_t
		leftPixel;						// current offset from start of active area where text is being drawn in this view
	int64_t
		topLine;						// current first line of edit universe that is displayed in the view
	uint64_t
		numPixels;						// number of pixels of "active" area in the view
	uint32_t
		numLines;						// number of lines that can be displayed
	uint32_t
		lineHeight;						// number of pixels per line vertically
	uint32_t
		lineAscent;						// tells where the baseline of text is for each line in the view
	bool
		blinkState;						// tells what state (true = on) (false = off) the cursor blink is currently in
	uint32_t
		nextBlinkTime;					// tells the next time that the cursor should change blink state
	uint64_t
		startPosition,					// position of the start of the line that's at the top (kept current for the view)
		endPosition;					// position of the start od the line just past the bottom
	int64_t
		startLine,						// used during text replacements to keep track of various lines
		endLine;
	uint64_t
		totalLinesAtStart;				// used during text replacement to see if the total number of lines changes
} PLATFORM_VIEW_INFO;

void ViewDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer,Region invalidRegion);
void ViewCursorTask(EDITOR_VIEW *view);
bool ViewTrack(EDITOR_VIEW *view,XEvent *event);
void ViewKeyEventHandle(EDITOR_VIEW *view,EDITOR_KEY *editorKey);
void ViewBoundsSet(EDITOR_VIEW *view,EDITOR_RECT *newBounds);
void PlatformViewTextXToPositionGet(EDITOR_VIEW *view,uint64_t linePosition,int64_t textX,uint64_t *betweenOffset,uint64_t *charOffset);
void PlatformViewTextPositionToXGet(EDITOR_VIEW *view,uint64_t position,int64_t *textX);
void PlatformViewCoordinateXToTextCoordinateX(EDITOR_VIEW *view,int64_t viewX,int64_t *textX);
void PlatformViewCoordinateLineToTextCoordinateLine(EDITOR_VIEW *view,int64_t viewLine,int64_t *textLine);
void PlatformViewCoordinateYToTextCoordinateLine(EDITOR_VIEW *view,int64_t viewY,int64_t *textLine);
void PlatformViewCoordinateToTextCoordinate(EDITOR_VIEW *view,int64_t viewX,int64_t viewY,int64_t *textX,int64_t *textLine);
void PlatformViewMetaChangeEnd(EDITOR_VIEW *view);
void PlatformViewMetaChangeStart(EDITOR_VIEW *view);
void PlatformViewStyleAreaChangeEnd(EDITOR_VIEW *view,STYLE_UNIVERSE *style);
void PlatformViewStyleAreaChangeStart(EDITOR_VIEW *view,STYLE_UNIVERSE *style);
void PlatformViewSelectionAreaChangeEnd(EDITOR_VIEW *view,SELECTION_UNIVERSE *selection);
void PlatformViewSelectionAreaChangeStart(EDITOR_VIEW *view,SELECTION_UNIVERSE *selection);
void PlatformViewColorSchemeChangeEnd(EDITOR_VIEW *view);
void PlatformViewColorSchemeChangeStart(EDITOR_VIEW *view);
void PlatformViewTextPieceReplaceEnd(EDITOR_VIEW *view,uint64_t position,uint64_t startLength,uint64_t endLength);
void PlatformViewTextPieceReplaceStart(EDITOR_VIEW *view,uint64_t position,uint64_t startLength,uint64_t endLength);
void PlatformViewTextReplaceEnd(EDITOR_VIEW *view);
void PlatformViewTextReplaceStart(EDITOR_VIEW *view);
void PlatformViewFocusChange(EDITOR_VIEW *view);
void PlatformViewScrollBoundsGet(EDITOR_VIEW *view,int64_t *leftPixel,int64_t *topLine,uint64_t *numPixels,uint64_t *numLines);
void PlatformViewActiveBoundsGet(EDITOR_VIEW *view,uint64_t *numPixels,uint64_t *numLines);
void PlatformViewOriginGet(EDITOR_VIEW *view,int64_t *leftPixel,int64_t *topLine);
void PlatformViewOriginSet(EDITOR_VIEW *view,int64_t newLeftPixel,int64_t newTopLine);
void PlatformViewInfoDispose(EDITOR_VIEW *view);
bool PlatformViewInfoCreate(EDITOR_VIEW *view,EDITOR_RECT *boundsRect);
