// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

typedef void WINDOW_CONFIGURE_FUNCTION(EDITOR_WINDOW *window);
typedef void WINDOW_UPDATE_FUNCTION(EDITOR_WINDOW *window);
typedef void WINDOW_EVENT_FUNCTION(EDITOR_WINDOW *window,XEvent *event);
typedef void WINDOW_PERIODIC_FUNCTION(EDITOR_WINDOW *window);

typedef struct editorWindow						// this structure is used to locate our data based on an x window ID
{
	uint16_t
		windowType;								// which type of window this is
	Window
		xWindow;								// x window id
	uint32_t
		w,h;									// width and height of the content area of the window (updated as window is configured)
	Region
		invalidRegion;							// the region that collects our invalidation information
	void
		*windowInfo;							// pointer to type-specific structure that keeps all info needed for a given type of window
	WINDOW_CONFIGURE_FUNCTION
		*configureFunction;						// tells what function to call after the window has been configured (resized) (NULL means none)
	WINDOW_UPDATE_FUNCTION
		*updateFunction;						// tells what function to call to update (redraw) the window
	WINDOW_EVENT_FUNCTION
		*eventFunction;							// tells what function to call to pass an event to the window
	WINDOW_PERIODIC_FUNCTION
		*periodicFunction;						// tells what function to call periodically when window has focus
	struct editorWindow
		*nextWindow,							// next one of our windows, or NULL if none
		*previousWindow;						// previous window, or NULL if none
} EDITOR_WINDOW;


void EditorDisplayDimensionsGet(uint32_t *width,uint32_t *height);
void WindowRectInvalidate(EDITOR_WINDOW *window,EDITOR_RECT *rect);
void WindowRegionInvalidate(EDITOR_WINDOW *window,Region region);
void InvalidWindowsRedraw();
void WindowClassHintsSet(EDITOR_WINDOW *window);
void WindowIconSet(EDITOR_WINDOW *window);
void WaitUntilMappedNoFocus(EDITOR_WINDOW *window);
void WaitUntilMapped(EDITOR_WINDOW *window);
void WaitUntilDestroyed(EDITOR_WINDOW *window);
void WindowElementLink(EDITOR_WINDOW *window);
void WindowElementUnlink(EDITOR_WINDOW *window);
EDITOR_WINDOW *WindowElementLocate(Window xWindow);
void EditorWindowMinimize(EDITOR_WINDOW *window);
void EditorWindowUnminimize(EDITOR_WINDOW *window);
void EditorTopWindowSet(EDITOR_WINDOW *window);
EDITOR_WINDOW *EditorTopWindowTypeGet(uint16_t windowType);
bool SortedEditorWindowTypeMaskListGet(uint16_t windowTypeMask,uint32_t *numElements,EDITOR_WINDOW ***list);
EDITOR_WINDOW *EditorFocusWindowGet();
void FocusAwayFrom(EDITOR_WINDOW *window);
void PlatformScreenOffsetAndDimensionsGet(int32_t *xOffset,int32_t *yOffset,uint32_t *width,uint32_t *height);
