// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// This include file is used to communicate with the platform.
// It contains all prototypes for porting functions that editor
// expects from the platform implementation.

const char *PlatformVersionGet();
void PlatformBeep();
void PlatformMessageReport(const char *format,...);
void PlatformBusyShow();
void PlatformBusyUnShow();
char *PlatformWindowTitleFromPathCreate(const char *absolutePath);
char *PlatformAbsolutePathCreate(const char *relativePath);
bool PlatformStartupScriptLocate(char *scriptPath);
bool PlatformFileRead(EDITOR_FILE *file,uint8_t *buffer,uint64_t bytesToRead,uint64_t *bytesRead);
bool PlatformFileWrite(EDITOR_FILE *file,uint8_t *buffer,uint64_t bytesToWrite,uint64_t *bytesWritten);
void PlatformFileClose(EDITOR_FILE *file);
EDITOR_FILE *PlatformFileReadOpen(const char *path);
EDITOR_FILE *PlatformFileWriteOpen(const char *path);

void PlatformScreenOffsetAndDimensionsGet(int32_t *xOffset,int32_t *yOffset,uint32_t *width,uint32_t *height);

void PlatformViewTextXToPositionGet(EDITOR_VIEW *view,uint64_t linePosition,int64_t textX,uint64_t *betweenOffset,uint64_t *charOffset);
void PlatformViewTextPositionToXGet(EDITOR_VIEW *view,uint64_t position,int64_t *textX);
void PlatformViewCoordinateXToTextCoordinateX(EDITOR_VIEW *view,int64_t viewX,int64_t *textX);
void PlatformViewCoordinateLineToTextCoordinateLine(EDITOR_VIEW *view,int64_t viewLine,int64_t *textLine);
void PlatformViewCoordinateYToTextCoordinateLine(EDITOR_VIEW *view,int64_t viewY,int64_t *textLine);
void PlatformViewCoordinateToTextCoordinate(EDITOR_VIEW *view,int64_t viewX,int64_t viewY,int64_t *textX,int64_t *textLine);
void PlatformViewMetaChangeEnd(EDITOR_VIEW *view);
void PlatformViewMetaChangeStart(EDITOR_VIEW *view);
void PlatformViewStyleAreaChangeEnd(EDITOR_VIEW *view,STYLE_UNIVERSE *style);
void PlatformViewStyleAreaChangeStart(EDITOR_VIEW *view,STYLE_UNIVERSE *style);
void PlatformViewSelectionAreaChangeEnd(EDITOR_VIEW *view,SELECTION_UNIVERSE *selection);
void PlatformViewSelectionAreaChangeStart(EDITOR_VIEW *view,SELECTION_UNIVERSE *selection);
void PlatformViewColorSchemeChangeEnd(EDITOR_VIEW *view);
void PlatformViewColorSchemeChangeStart(EDITOR_VIEW *view);
void PlatformViewTextPieceReplaceEnd(EDITOR_VIEW *view,uint64_t position,uint64_t startLength,uint64_t endLength);
void PlatformViewTextPieceReplaceStart(EDITOR_VIEW *view,uint64_t position,uint64_t startLength,uint64_t endLength);
void PlatformViewTextReplaceEnd(EDITOR_VIEW *view);
void PlatformViewTextReplaceStart(EDITOR_VIEW *view);
void PlatformViewFocusChange(EDITOR_VIEW *view);
void PlatformViewScrollBoundsGet(EDITOR_VIEW *view,int64_t *leftPixel,int64_t *topLine,uint64_t *numPixels,uint64_t *numLines);
void PlatformViewActiveBoundsGet(EDITOR_VIEW *view,uint64_t *numPixels,uint64_t *numLines);
void PlatformViewOriginGet(EDITOR_VIEW *view,int64_t *leftPixel,int64_t *topLine);
void PlatformViewOriginSet(EDITOR_VIEW *view,int64_t newLeftPixel,int64_t newTopLine);
void PlatformViewInfoDispose(EDITOR_VIEW *view);
bool PlatformViewInfoCreate(EDITOR_VIEW *view,EDITOR_RECT *boundsRect);

bool PlatformColorNameToColor(const char *colorName,EDITOR_COLOR *color);

void PlatformDocumentWindowUnminimize(EDITOR_WINDOW *window);
void PlatformDocumentWindowMinimize(EDITOR_WINDOW *window);
EDITOR_BUFFER *PlatformDocumentWindowBufferGet(EDITOR_WINDOW *window);
EDITOR_WINDOW *PlatformDocumentWindowLocate(EDITOR_BUFFER *buffer);
EDITOR_WINDOW *PlatformDocumentWindowActiveGet();
EDITOR_WINDOW *PlatformDocumentWindowTopGet();
void PlatformDocumentWindowTopSet(EDITOR_WINDOW *window);
bool PlatformDocumentWindowSortedListGet(uint32_t *numElements,EDITOR_WINDOW ***list);
EDITOR_VIEW *PlatformDocumentWindowCurrentViewGet(EDITOR_WINDOW *window);
bool PlatformDocumentWindowTitleGet(EDITOR_WINDOW *window,char *title,uint32_t stringBytes);
bool PlatformDocumentWindowTitleSet(EDITOR_WINDOW *window,const char *title);
bool PlatformDocumentWindowRectGet(EDITOR_WINDOW *window,EDITOR_RECT *rect);
bool PlatformDocumentWindowRectSet(EDITOR_WINDOW *window,EDITOR_RECT *rect);
void PlatformDocumentWindowClose(EDITOR_WINDOW *window);
EDITOR_WINDOW *PlatformDocumentWindowOpen(EDITOR_BUFFER *buffer,EDITOR_RECT *rect,const char *title,bool useUtf8,uint32_t tabSize,const char *fontName,const char *colorSchemeName);

EDITOR_MENU *PlatformMenuRelatedGet(EDITOR_MENU *menu,int menuRelationship);
void PlatformMenuActiveSet(EDITOR_MENU *menu,bool active);
char *PlatformMenuNameGet(EDITOR_MENU *menu);
bool PlatformMenuNameSet(EDITOR_MENU *menu,const char *name);
char *PlatformMenuAttributesGet(EDITOR_MENU *menu);
bool PlatformMenuAttributesSet(EDITOR_MENU *menu,const char *attributes);
char *PlatformMenuDataTextGet(EDITOR_MENU *menu);
bool PlatformMenuDataTextSet(EDITOR_MENU *menu,const char *dataText);
bool PlatformMenuGet(int argc,const char *argv[],EDITOR_MENU **menu);
void PlatformMenuDispose(EDITOR_MENU *menu);
EDITOR_MENU *PlatformMenuCreate(int argc,const char *argv[],int menuRelationship,const char *name,const char *attributes,const char *dataText,bool active);

uint32_t PlatformRemapEditorModifiers(uint32_t modifiersIn);
bool PlatformKeyNameToKeyCode(const char *name,uint32_t *keyCode);
const char *PlatformKeyCodeToKeyName(uint32_t keyCode);
bool PlatformKeyPressGet(EDITOR_KEY *editorKey,bool wait,bool clearBuffered);
void PlatformEventLoop(int argc,char *argv[]);
void PlatformQuit();
void PlatformEventsAbortClear();
bool PlatformEventsAbortCheck();
uint32_t PlatformMonotonicTime();

bool PlatformClipboardImport();
bool PlatformClipboardExport();

void PlatformFontCompositeNameSplit(const char *compositeName,char *familyName,uint32_t *size,bool *bold,bool *oblique);
void PlatformFontComponentNameJoin(const char *familyName,uint32_t size,bool bold,bool oblique,char *compositeName);
const char *PlatformFontCompositeNameGet(EDITOR_FONT *font);
void PlatformFontFree(EDITOR_FONT *font);
EDITOR_FONT *PlatformFontLoad(const char *compositeName);
EDITOR_FONT *PlatformFontDuplicate(EDITOR_FONT *original);

bool PlatformDialogOk(const char *text);
bool PlatformDialogOkCancel(const char *text,bool *cancel);
bool PlatformDialogYesNoCancel(const char *text,bool *yes,bool *cancel);
bool PlatformDialogSimpleText(const char *title,char *enteredText,uint32_t stringBytes,bool *cancel);
bool PlatformDialogSearchReplace(EDITOR_BUFFER *searchBuffer,EDITOR_BUFFER *replaceBuffer,bool *backwards,bool *wrapAround,bool *ignoreCase,bool *limitScope,bool *selectionExpr,bool *useSubexpression,uint32_t *subexpressionIndex,bool *replaceFunction,uint16_t *searchType,bool *cancel);
bool PlatformDialogSimpleList(const char *title,uint32_t numElements,const char **listElements,bool *selectedElements,bool *cancel);
bool PlatformDialogOpenFile(const char *title,char *fullPath,uint32_t stringBytes,char ***listElements,bool *cancel);
void PlatformDialogFreeOpenFilePaths(char **paths);
bool PlatformDialogSaveFile(const char *title,char *fullPath,uint32_t stringBytes,bool *cancel);
bool PlatformDialogChoosePath(const char *title,char *fullPath,uint32_t stringBytes,bool *cancel);
bool PlatformDialogChooseFont(const char *title,char *font,uint32_t stringBytes,bool *cancel);
void PlatformDialogCenterPointSet(int32_t centerX,int32_t centerY);

bool PlatformTaskDataRead(EDITOR_TASK *task,uint8_t *buffer,uint64_t bufferLength,uint64_t *bytesRead);
bool PlatformTaskStatus(EDITOR_TASK *task,const char **terminationString);
bool PlatformTaskEofWrite(EDITOR_TASK *task);
bool PlatformTaskDataWrite(EDITOR_TASK *task,const uint8_t *data,uint64_t numBytes);
const char *PlatformTaskCommandLineGet(EDITOR_TASK *task);
void PlatformTaskKill(EDITOR_TASK *task);
void PlatformTaskDisconnect(EDITOR_TASK *task);
EDITOR_TASK *PlatformTaskSpawn(const uint8_t *commandLine);

void *MNewPtr(uint64_t size);
void *MNewPtrClr(uint64_t size);
void MDisposePtr(void *pointer);
void MMoveMem(const void *source,void *dest,uint64_t numBytes);
uint64_t MGetNumAllocatedPointers();

bool PlatformShellCommandsAdd(Tcl_Interp *interpreter);

void PlatformLateUnInit();
bool PlatformLateInit();
void PlatformUnInit();
bool PlatformInit();
void PlatformEarlyUnInit();
bool PlatformEarlyInit();
