// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

typedef struct sparseArrayCacheEntry
{
	uint64_t
		cArrayIndex;				// the number of carray elements to the cached sparse record (0 if no cache)
	uint64_t
		sparseIndex;				// the number of sparse elements to the cached sparse record
} SPARSE_ARRAY_CACHE_ENTRY;

typedef struct sparseArrayHeader
{
	uint64_t
		offset;						// offset to beginning of element from end of previous element
	uint64_t
		length;						// number of elements taken by this entry
} SPARSE_ARRAY_HEADER;

typedef struct sparseArray
{
	CARRAY
		*elementCArray;				// array of elements
	SPARSE_ARRAY_CACHE_ENTRY
		cacheEntry;					// one entry in a cache that tracks the last sparse element accessed
} SPARSE_ARRAY;

void SparseArraySanity(const SPARSE_ARRAY *sparseArray);
void SparseArrayRangeDelete(SPARSE_ARRAY *sparseArray,uint64_t startPosition,uint64_t numElements);
bool SparseArrayRangeInsert(SPARSE_ARRAY *sparseArray,uint64_t startPosition,uint64_t numElements);
bool SparseArrayRecordForPositionGet(SPARSE_ARRAY *sparseArray,uint64_t position,uint64_t *startPosition,uint64_t *numElements,void **record);
bool SparseArrayRecordAtOrAfterPositionGet(SPARSE_ARRAY *sparseArray,uint64_t position,uint64_t *startPosition,uint64_t *numElements,void **record);
bool SparseArrayRecordAtOrBeforePositionGet(SPARSE_ARRAY *sparseArray,uint64_t position,uint64_t *startPosition,uint64_t *numElements,void **record);
bool SparseArrayRangeClear(SPARSE_ARRAY *sparseArray,uint64_t startPosition,uint64_t numElements);
bool SparseArrayRangeSet(SPARSE_ARRAY *sparseArray,uint64_t startPosition,uint64_t numElements,void *value);
bool SparseArrayElementEndsGet(SPARSE_ARRAY *sparseArray,uint64_t *startElementOffset,uint64_t *endElementOffset);
uint64_t SparseArrayTotalOccupiedElementsGet(SPARSE_ARRAY *sparseArray);
uint64_t SparseArrayTotalElementsGet(SPARSE_ARRAY *sparseArray);
uint64_t SparseArraySegmentsGet(SPARSE_ARRAY *sparseArray);
bool SparseArrayIsEmpty(SPARSE_ARRAY *sparseArray);
void SparseArrayUnInit(SPARSE_ARRAY *sparseArray);
bool SparseArrayInit(SPARSE_ARRAY *sparseArray,uint64_t elementSize);
